# This file is part of PyEMMA.
#
# Copyright (c) 2015, 2014 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as _np
from msmtools import estimation as msmest

from pyemma.msm.estimators._msm_estimator_base import _MSMEstimator
from pyemma.util.annotators import aliased, fix_docs
from pyemma.util import types as _types

__all__ = ['MaximumLikelihoodMSM']


@fix_docs
@aliased
class MaximumLikelihoodMSM(_MSMEstimator):
    r"""Maximum likelihood estimator for MSMs given discrete trajectory statistics"""
    __serialize_fields = ('_C_active', '_C_full',
                          '_full2active', '_connected_sets',
                          '_nstates', '_nstates_full',
                          )
    __serialize_version = 0

    def __init__(self, lag=1, reversible=True, statdist_constraint=None,
                 count_mode='sliding', sparse=False,
                 connectivity='largest', dt_traj='1 step', maxiter=1000000,
                 maxerr=1e-8, score_method='VAMP2', score_k=10,
                 mincount_connectivity='1/n'):
        r"""Maximum likelihood estimator for MSMs given discrete trajectory statistics

        Parameters
        ----------
        lag : int
            lag time at which transitions are counted and the transition matrix is
            estimated.

        reversible : bool, optional, default = True
            If true compute reversible MSM, else non-reversible MSM

        statdist : (M,) ndarray, optional
            Stationary vector on the full set of states. Estimation will be
            made such the the resulting transition matrix has this distribution
            as an equilibrium distribution. Set probabilities to zero if these
            states should be excluded from the analysis.

        count_mode : str, optional, default='sliding'
            mode to obtain count matrices from discrete trajectories. Should be
            one of:

            * 'sliding' : A trajectory of length T will have :math:`T-tau` counts
              at time indexes

              .. math::

                 (0 \rightarrow \tau), (1 \rightarrow \tau+1), ..., (T-\tau-1 \rightarrow T-1)

            * 'effective' : Uses an estimate of the transition counts that are
              statistically uncorrelated. Recommended when used with a
              Bayesian MSM.
            * 'sample' : A trajectory of length T will have :math:`T/tau` counts
              at time indexes

              .. math::

                    (0 \rightarrow \tau), (\tau \rightarrow 2 \tau), ..., (((T/tau)-1) \tau \rightarrow T)

        sparse : bool, optional, default = False
            If true compute count matrix, transition matrix and all derived
            quantities using sparse matrix algebra. In this case python sparse
            matrices will be returned by the corresponding functions instead of
            numpy arrays. This behavior is suggested for very large numbers of
            states (e.g. > 4000) because it is likely to be much more efficient.
        connectivity : str, optional, default = 'largest'
            Connectivity mode. Three methods are intended (currently only 'largest'
            is implemented)

            * 'largest' : The active set is the largest reversibly connected set.
              All estimation will be done on this subset and all quantities
              (transition matrix, stationary distribution, etc) are only defined
              on this subset and are correspondingly smaller than the full set
              of states
            * 'all' : The active set is the full set of states. Estimation will be
              conducted on each reversibly connected set separately. That means
              the transition matrix will decompose into disconnected submatrices,
              the stationary vector is only defined within subsets, etc.
              Currently not implemented.
            * 'none' : The active set is the full set of states. Estimation will
              be conducted on the full set of
              states without ensuring connectivity. This only permits
              nonreversible estimation. Currently not implemented.

        dt_traj : str, optional, default='1 step'
            Description of the physical time of the input trajectories. May be used
            by analysis algorithms such as plotting tools to pretty-print the axes.
            By default '1 step', i.e. there is no physical time unit. Specify by a
            number, whitespace and unit. Permitted units are (* is an arbitrary
            string):

            |  'fs',  'femtosecond*'
            |  'ps',  'picosecond*'
            |  'ns',  'nanosecond*'
            |  'us',  'microsecond*'
            |  'ms',  'millisecond*'
            |  's',   'second*'

        maxiter: int, optioanl, default = 1000000
            Optional parameter with reversible = True. maximum number of iterations
            before the transition matrix estimation method exits
        maxerr : float, optional, default = 1e-8
            Optional parameter with reversible = True.
            convergence tolerance for transition matrix estimation.
            This specifies the maximum change of the Euclidean norm of relative
            stationary probabilities (:math:`x_i = \sum_k x_{ik}`). The relative
            stationary probability changes
            :math:`e_i = (x_i^{(1)} - x_i^{(2)})/(x_i^{(1)} + x_i^{(2)})` are used
            in order to track changes in small probabilities. The Euclidean norm
            of the change vector, :math:`|e_i|_2`, is compared to maxerr.

        score_method : str, optional, default='VAMP2'
            Score to be used with score function. Available are:

            |  'VAMP1'  [1]_
            |  'VAMP2'  [1]_
            |  'VAMPE'  [1]_

        score_k : int or None
            The maximum number of eigenvalues or singular values used in the
            score. If set to None, all available eigenvalues will be used.

        mincount_connectivity : float or '1/n'
            minimum number of counts to consider a connection between two states.
            Counts lower than that will count zero in the connectivity check and
            may thus separate the resulting transition matrix. The default
            evaluates to 1/nstates.

        References
        ----------
        .. [1] H. Wu and F. Noe: Variational approach for learning Markov processes from time series data
            (in preparation)

        """
        super(MaximumLikelihoodMSM, self).__init__(lag=lag, reversible=reversible, count_mode=count_mode,
                                                   sparse=sparse, connectivity=connectivity, dt_traj=dt_traj,
                                                   score_method=score_method, score_k=score_k,
                                                   mincount_connectivity=mincount_connectivity)

        self.statdist_constraint = _types.ensure_ndarray_or_None(statdist_constraint, ndim=None, kind='numeric')
        if self.statdist_constraint is not None:  # renormalize
            self.statdist_constraint /= self.statdist_constraint.sum()

        # convergence parameters
        self.maxiter = maxiter
        self.maxerr = maxerr

    def _prepare_input_revpi(self, C, pi):
        """Max. state index visited by trajectories"""
        nC = C.shape[0]
        # Max. state index of the stationary vector array
        npi = pi.shape[0]
        # pi has to be defined on all states visited by the trajectories
        if nC > npi:
            raise ValueError('There are visited states for which no stationary probability is given')
        # Reduce pi to the visited set
        pi_visited = pi[0:nC]
        # Find visited states with positive stationary probabilities"""
        pos = _np.where(pi_visited > 0.0)[0]
        # Reduce C to positive probability states"""
        C_pos = msmest.connected_cmatrix(C, lcc=pos)
        if C_pos.sum() == 0.0:
            errstr = """The set of states with positive stationary
            probabilities is not visited by the trajectories. A MSM
            reversible with respect to the given stationary vector can
            not be estimated"""
            raise ValueError(errstr)
        # Compute largest connected set of C_pos, undirected connectivity"""
        lcc = msmest.largest_connected_set(C_pos, directed=False)
        return pos[lcc]

    def _estimate(self, dtrajs):
        """ Estimates the MSM """
        # get trajectory counts. This sets _C_full and _nstates_full
        dtrajstats = self._get_dtraj_stats(dtrajs)
        self._C_full = dtrajstats.count_matrix()  # full count matrix
        self._nstates_full = self._C_full.shape[0]  # number of states

        # set active set. This is at the same time a mapping from active to full
        if self.connectivity == 'largest':
            if self.statdist_constraint is None:
                # statdist not given - full connectivity on all states
                self.active_set = dtrajstats.largest_connected_set
            else:
                active_set = self._prepare_input_revpi(self._C_full,
                                                       self.statdist_constraint)
                self.active_set = active_set
        else:
            # for 'None' and 'all' all visited states are active
            self.active_set = dtrajstats.visited_set

        # FIXME: setting is_estimated before so that we can start using the parameters just set, but this is not clean!
        # is estimated
        self._is_estimated = True

        # if active set is empty, we can't do anything.
        if _np.size(self.active_set) == 0:
            raise RuntimeError('Active set is empty. Cannot estimate MSM.')

        # active count matrix and number of states
        self._C_active = dtrajstats.count_matrix(subset=self.active_set)

        # continue sparse or dense?
        if not self.sparse:
            # converting count matrices to arrays. As a result the
            # transition matrix and all subsequent properties will be
            # computed using dense arrays and dense matrix algebra.
            self._C_full = self._C_full.toarray()
            self._C_active = self._C_active.toarray()

        self._nstates = self._C_active.shape[0]

        # computed derived quantities
        # back-mapping from full to lcs
        self._full2active = -1 * _np.ones(dtrajstats.nstates, dtype=int)
        self._full2active[self.active_set] = _np.arange(len(self.active_set))

        # restrict stationary distribution to active set
        if self.statdist_constraint is None:
            statdist_active = None
        else:
            statdist_active = self.statdist_constraint[self.active_set]
            statdist_active /= statdist_active.sum()  # renormalize

        opt_args = {}
        # TODO: non-rev estimate of msmtools does not comply with its own api...
        if statdist_active is None and self.reversible:
            opt_args['return_statdist'] = True

        # Estimate transition matrix
        if self.connectivity == 'largest':
            P = msmest.transition_matrix(self._C_active, reversible=self.reversible,
                                         mu=statdist_active, maxiter=self.maxiter,
                                         maxerr=self.maxerr, **opt_args)
        elif self.connectivity == 'none':
            # reversible mode only possible if active set is connected
            # - in this case all visited states are connected and thus
            # this mode is identical to 'largest'
            if self.reversible and not msmest.is_connected(self._C_active):
                raise ValueError('Reversible MSM estimation is not possible with connectivity mode "none", '
                                 'because the set of all visited states is not reversibly connected')
            P = msmest.transition_matrix(self._C_active, reversible=self.reversible,
                                         mu=statdist_active,
                                         maxiter=self.maxiter, maxerr=self.maxerr,
                                         **opt_args
                                         )
        else:
            raise NotImplementedError(
                'MSM estimation with connectivity=%s is currently not implemented.' % self.connectivity)

        # msmtools returns a tuple for statdist_active = None.
        if isinstance(P, tuple):
            P, statdist_active = P

        # Done. We set our own model parameters, so this estimator is
        # equal to the estimated model.
        self._dtrajs_full = dtrajs
        self._connected_sets = dtrajstats.connected_sets
        self.set_model_params(P=P, pi=statdist_active, reversible=self.reversible,
                              dt_model=self.timestep_traj.get_scaled(self.lag))

        return self

    # TODO: change to statistically effective count matrix!
    @property
    def effective_count_matrix(self):
        """Statistically uncorrelated transition counts within the active set of states

        You can use this count matrix for Bayesian estimation or error perturbation.

        References
        ----------
        [1] Noe, F. (2015) Statistical inefficiency of Markov model count matrices
            http://publications.mi.fu-berlin.de/1699/1/autocorrelation_counts.pdf

        """
        self._check_is_estimated()
        Ceff_full = msmest.effective_count_matrix(self._dtrajs_full, self.lag)
        from pyemma.util.linalg import submatrix
        Ceff = submatrix(Ceff_full, self.active_set)
        return Ceff
        # return self._C_active / float(self.lag)



import unittest
import numpy as np
from .. import moments

__author__ = 'noe'

class TestMoments(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.X_2 = np.random.rand(10000, 2)
        cls.Y_2 = np.random.rand(10000, 2)
        # dense data
        cls.X_10 = np.random.rand(10000, 10)
        cls.Y_10 = np.random.rand(10000, 10)
        cls.X_100 = np.random.rand(10000, 100)
        cls.Y_100 = np.random.rand(10000, 100)
        # sparse zero data
        cls.X_10_sparsezero = np.zeros((10000, 10))
        cls.X_10_sparsezero[:, 0] = cls.X_10[:, 0]
        cls.Y_10_sparsezero = np.zeros((10000, 10))
        cls.Y_10_sparsezero[:, 0] = cls.Y_10[:, 0]
        cls.X_100_sparsezero = np.zeros((10000, 100))
        cls.X_100_sparsezero[:, :10] = cls.X_100[:, :10]
        cls.Y_100_sparsezero = np.zeros((10000, 100))
        cls.Y_100_sparsezero[:, :10] = cls.Y_100[:, :10]
        # sparse const data
        cls.X_10_sparseconst = np.ones((10000, 10))
        cls.X_10_sparseconst[:, 0] = cls.X_10[:, 0]
        cls.Y_10_sparseconst = 2*np.ones((10000, 10))
        cls.Y_10_sparseconst[:, 0] = cls.Y_10[:, 0]
        cls.X_100_sparseconst = np.ones((10000, 100))
        cls.X_100_sparseconst[:, :10] = cls.X_100[:, :10]
        cls.Y_100_sparseconst = 2*np.ones((10000, 100))
        cls.Y_100_sparseconst[:, :10] = cls.Y_100[:, :10]
        # boolean data
        cls.Xb_2 = np.random.randint(0, 2, size=(10000, 2))
        cls.Xb_2 = cls.Xb_2.astype(np.bool)
        cls.Xb_10 = np.random.randint(0, 2, size=(10000, 10))
        cls.Xb_10 = cls.Xb_10.astype(np.bool)
        cls.Xb_10_sparsezero = np.zeros((10000, 10), dtype=np.bool)
        cls.Xb_10_sparsezero[:, 0] = cls.Xb_10[:, 0]
        # generate weights:
        cls.weights = np.random.rand(10000)
        # Set the lag time for time-lagged tests:
        cls.lag = 50
        # column subsets
        cls.cols_2 = np.array([0])
        cls.cols_10 = np.random.choice(10, 5, replace=False)
        cls.cols_100 = np.random.choice(100, 20, replace=False)
        # sparse tolerance
        cls.sparse_tol = 1e-14

        return cls

    def _test_moments_X(self, X, column_selection, remove_mean=False, sparse_mode='auto', sparse_tol=0.0, weights=None):
        # proposed solution
        w, s_X, C_XX = moments.moments_XX(X, remove_mean=remove_mean, modify_data=False,
                                          sparse_mode=sparse_mode, sparse_tol=sparse_tol,
                                          weights=weights)
        # reference
        X = X.astype(np.float64)
        if weights is not None:
            X1 = weights[:, None] * X
            w = weights.sum()
        else:
            X1 = X
            w = X.shape[0]
        s_X_ref = X1.sum(axis=0)
        if remove_mean:
            X = X - (1.0 / w) * s_X_ref
        if weights is not None:
            X1 = weights[:, None] * X
        else:
            X1 = X
        C_XX_ref = np.dot(X1.T, X)
        # test
        assert np.allclose(s_X, s_X_ref)
        assert np.allclose(C_XX, C_XX_ref)
        # column subsets
        w, s_X, C_XX = moments.moments_XX(X, remove_mean=remove_mean, modify_data=False,
                                          sparse_mode=sparse_mode, sparse_tol=sparse_tol,
                                          weights=weights, column_selection=column_selection)
        assert np.allclose(C_XX, C_XX_ref[:, column_selection])
        # diagonal only
        if sparse_mode is not 'sparse':
            w, s_X, C_XX = moments.moments_XX(X, remove_mean=remove_mean, modify_data=False,
                                              sparse_mode=sparse_mode, weights=weights, diag_only=True)
            assert np.allclose(C_XX, np.diag(C_XX_ref))

    def test_moments_X(self):
        # simple test, dense
        self._test_moments_X(self.X_10, self.cols_10, remove_mean=False, sparse_mode='dense')
        self._test_moments_X(self.X_100, self.cols_100, remove_mean=False, sparse_mode='dense')
        # mean-free, dense
        self._test_moments_X(self.X_10, self.cols_10, remove_mean=True, sparse_mode='dense')
        self._test_moments_X(self.X_100, self.cols_100, remove_mean=True, sparse_mode='dense')
        # weighted test, simple, dense:
        self._test_moments_X(self.X_10, self.cols_10, remove_mean=False, sparse_mode='dense', weights=self.weights)
        self._test_moments_X(self.X_100, self.cols_100, remove_mean=False, sparse_mode='dense', weights=self.weights)
        # weighted test, mean-free, dense:
        self._test_moments_X(self.X_10, self.cols_10, remove_mean=True, sparse_mode='dense', weights=self.weights)
        self._test_moments_X(self.X_100, self.cols_100, remove_mean=True, sparse_mode='dense', weights=self.weights)

    def test_moments_X_sparsezero(self):
        # simple test, sparse
        self._test_moments_X(self.X_10_sparsezero, self.cols_10, remove_mean=False, sparse_mode='sparse', sparse_tol=self.sparse_tol)
        self._test_moments_X(self.X_100_sparsezero, self.cols_100, remove_mean=False, sparse_mode='sparse', sparse_tol=self.sparse_tol)
        # mean-free, sparse
        self._test_moments_X(self.X_10_sparsezero, self.cols_10, remove_mean=True, sparse_mode='sparse', sparse_tol=self.sparse_tol)
        self._test_moments_X(self.X_100_sparsezero, self.cols_100, remove_mean=True, sparse_mode='sparse', sparse_tol=self.sparse_tol)
        # weighted, sparse
        self._test_moments_X(self.X_10_sparsezero, self.cols_10, remove_mean=False, sparse_mode='sparse', sparse_tol=self.sparse_tol, weights=self.weights)
        self._test_moments_X(self.X_100_sparsezero, self.cols_100, remove_mean=False, sparse_mode='sparse', sparse_tol=self.sparse_tol, weights=self.weights)
        # weighted, mean-free, sparse
        self._test_moments_X(self.X_10_sparsezero, self.cols_10, remove_mean=True, sparse_mode='sparse', sparse_tol=self.sparse_tol, weights=self.weights)
        self._test_moments_X(self.X_100_sparsezero, self.cols_100, remove_mean=True, sparse_mode='sparse', sparse_tol=self.sparse_tol, weights=self.weights)

    def test_moments_X_sparseconst(self):
        # simple test, sparse
        self._test_moments_X(self.X_10_sparseconst, self.cols_10, remove_mean=False, sparse_mode='sparse', sparse_tol=self.sparse_tol)
        self._test_moments_X(self.X_100_sparseconst, self.cols_100, remove_mean=False, sparse_mode='sparse', sparse_tol=self.sparse_tol)
        # mean-free, sparse
        self._test_moments_X(self.X_10_sparseconst, self.cols_10, remove_mean=True, sparse_mode='sparse', sparse_tol=self.sparse_tol)
        self._test_moments_X(self.X_100_sparseconst, self.cols_100, remove_mean=True, sparse_mode='sparse', sparse_tol=self.sparse_tol)
        # weighted, sparse:
        self._test_moments_X(self.X_10_sparseconst, self.cols_10, remove_mean=False, sparse_mode='sparse', sparse_tol=self.sparse_tol, weights=self.weights)
        self._test_moments_X(self.X_100_sparseconst, self.cols_100, remove_mean=False, sparse_mode='sparse', sparse_tol=self.sparse_tol, weights=self.weights)
        # weighted, mean-free, sparse:
        self._test_moments_X(self.X_10_sparseconst, self.cols_10, remove_mean=True, sparse_mode='sparse', sparse_tol=self.sparse_tol, weights=self.weights)
        self._test_moments_X(self.X_100_sparseconst, self.cols_100, remove_mean=True, sparse_mode='sparse', sparse_tol=self.sparse_tol, weights=self.weights)

    def test_boolean_moments(self):
        # standard tests
        self._test_moments_X(self.Xb_10, self.cols_10, remove_mean=False, sparse_mode='dense')
        self._test_moments_X(self.Xb_10, self.cols_10, remove_mean=True, sparse_mode='dense')
        self._test_moments_X(self.Xb_10_sparsezero, self.cols_10, remove_mean=False, sparse_mode='sparse', sparse_tol=self.sparse_tol)
        self._test_moments_X(self.Xb_10_sparsezero, self.cols_10, remove_mean=True, sparse_mode='sparse', sparse_tol=self.sparse_tol)
        # test integer recovery
        Cxx_ref = np.dot(self.Xb_10.astype(np.int64).T, self.Xb_10.astype(np.int64))  # integer
        s_X_ref = np.sum(self.Xb_10, axis=0)
        w, s_X, Cxx = moments.moments_XX(self.Xb_10, remove_mean=False, modify_data=False, sparse_mode='dense')
        s_X = np.round(s_X).astype(np.int64)
        Cxx = np.round(Cxx).astype(np.int64)
        assert np.array_equal(s_X, s_X_ref)
        assert np.array_equal(Cxx, Cxx_ref)


    def _test_moments_XY(self, X, Y, column_selection, symmetrize=False, remove_mean=False, sparse_mode='auto', sparse_tol=0.0, weights=None):
        w1, s_X, s_Y, C_XX, C_XY = moments.moments_XXXY(X, Y, remove_mean=remove_mean, modify_data=False,
                                                        symmetrize=symmetrize, sparse_mode=sparse_mode,
                                                        sparse_tol=sparse_tol, weights=weights)
        # reference
        T = X.shape[0]
        if weights is not None:
            X1 = weights[:, None] * X
            Y1 = weights[:, None] * Y
        else:
            X1 = X
            Y1 = Y
        s_X_ref = X1.sum(axis=0)
        s_Y_ref = Y1.sum(axis=0)
        if symmetrize:
            s_X_ref = s_X_ref + s_Y_ref
            s_Y_ref = s_X_ref
            if weights is not None:
                w = 2 * np.sum(weights)
            else:
                w = 2 * T
        else:
            if weights is not None:
                w = np.sum(weights)
            else:
                w = T
        if remove_mean:
            X = X - s_X_ref/float(w)
            Y = Y - s_Y_ref/float(w)
        if weights is not None:
            X1 = weights[:, None] * X
            Y1 = weights[:, None] * Y
        else:
            X1 = X
            Y1 = Y
        if symmetrize:
            C_XX_ref = np.dot(X1.T, X) + np.dot(Y1.T, Y)
            C_XY_ref = np.dot(X1.T, Y) + np.dot(Y1.T, X)
        else:
            C_XX_ref = np.dot(X1.T, X)
            C_XY_ref = np.dot(X1.T, Y)
        # test
        assert np.allclose(w1, w)
        assert np.allclose(s_X, s_X_ref)
        assert np.allclose(s_Y, s_Y_ref)
        assert np.allclose(C_XX, C_XX_ref)
        assert np.allclose(C_XY, C_XY_ref)
        # column subsets
        w1, s_X, s_Y, C_XX, C_XY = moments.moments_XXXY(X, Y, remove_mean=remove_mean, modify_data=False,
                                                       symmetrize=symmetrize, sparse_mode=sparse_mode, sparse_tol=sparse_tol,
                                                       weights=weights, column_selection=column_selection)
        assert np.allclose(C_XX, C_XX_ref[:, column_selection])
        assert np.allclose(C_XY, C_XY_ref[:, column_selection])
        # diagonal only
        if sparse_mode is not 'sparse' and X.shape[1] == Y.shape[1]:
            w1, s_X, s_Y, C_XX, C_XY = moments.moments_XXXY(X, Y, remove_mean=remove_mean, modify_data=False,
                                                           symmetrize=symmetrize, sparse_mode=sparse_mode, sparse_tol=sparse_tol,
                                                           weights=weights, diag_only=True)
            assert np.allclose(C_XX, np.diag(C_XX_ref))
            assert np.allclose(C_XY, np.diag(C_XY_ref))

    def test_moments_XY(self):
        # simple test, dense
        self._test_moments_XY(self.X_10, self.Y_10, self.cols_10, symmetrize=False, remove_mean=False, sparse_mode='dense')
        self._test_moments_XY(self.X_100, self.Y_10, self.cols_10, symmetrize=False, remove_mean=False, sparse_mode='dense')
        self._test_moments_XY(self.X_100, self.Y_100, self.cols_100, symmetrize=False, remove_mean=False, sparse_mode='dense')
        # mean-free, dense
        self._test_moments_XY(self.X_10, self.Y_10, self.cols_10, symmetrize=False, remove_mean=True, sparse_mode='dense')
        self._test_moments_XY(self.X_100, self.Y_10, self.cols_10, symmetrize=False, remove_mean=True, sparse_mode='dense')
        self._test_moments_XY(self.X_100, self.Y_100, self.cols_100, symmetrize=False, remove_mean=True, sparse_mode='dense')

    def test_moments_XY_weighted(self):
        # weighted test, dense
        self._test_moments_XY(self.X_10, self.Y_10, self.cols_10, symmetrize=False, remove_mean=False,
                              sparse_mode='dense', weights=self.weights)
        self._test_moments_XY(self.X_100, self.Y_100, self.cols_100, symmetrize=False, remove_mean=False,
                              sparse_mode='dense', weights=self.weights)
        # weighted test, mean-free, dense
        self._test_moments_XY(self.X_10, self.Y_10, self.cols_10, symmetrize=False, remove_mean=True,
                              sparse_mode='dense', weights=self.weights)
        self._test_moments_XY(self.X_100, self.Y_100, self.cols_100, symmetrize=False, remove_mean=True,
                              sparse_mode='dense', weights=self.weights)

    def test_moments_XY_sym(self):
        # simple test, dense, symmetric
        self._test_moments_XY(self.X_2, self.Y_2, self.cols_2, symmetrize=True, remove_mean=False, sparse_mode='dense')
        self._test_moments_XY(self.X_10, self.Y_10, self.cols_10, symmetrize=True, remove_mean=False, sparse_mode='dense')
        self._test_moments_XY(self.X_100, self.Y_100, self.cols_100, symmetrize=True, remove_mean=False, sparse_mode='dense')
        # mean-free, dense, symmetric
        self._test_moments_XY(self.X_2, self.Y_2, self.cols_2, symmetrize=True, remove_mean=True, sparse_mode='dense')
        self._test_moments_XY(self.X_10, self.Y_10, self.cols_10, symmetrize=True, remove_mean=True, sparse_mode='dense')
        self._test_moments_XY(self.X_100, self.Y_100, self.cols_100, symmetrize=True, remove_mean=True, sparse_mode='dense')

    def test_moments_XY_weighted_sym(self):
        # simple test, dense, symmetric
        self._test_moments_XY(self.X_2, self.Y_2, self.cols_2, symmetrize=True, remove_mean=False, sparse_mode='dense',
                              weights=self.weights)
        self._test_moments_XY(self.X_10, self.Y_10, self.cols_10, symmetrize=True, remove_mean=False, sparse_mode='dense'
                              , weights=self.weights)
        self._test_moments_XY(self.X_100, self.Y_100, self.cols_100, symmetrize=True, remove_mean=False, sparse_mode='dense',
                              weights=self.weights)
        # mean-free, dense, symmetric
        self._test_moments_XY(self.X_2, self.Y_2, self.cols_2, symmetrize=True, remove_mean=True, sparse_mode='dense',
                              weights=self.weights)
        self._test_moments_XY(self.X_10, self.Y_10, self.cols_10, symmetrize=True, remove_mean=True, sparse_mode='dense',
                              weights=self.weights)
        self._test_moments_XY(self.X_100, self.Y_100, self.cols_100, symmetrize=True, remove_mean=True, sparse_mode='dense',
                              weights=self.weights)

    def test_moments_XY_sparsezero(self):
        # simple test, sparse
        self._test_moments_XY(self.X_10_sparsezero, self.Y_10_sparsezero, self.cols_10, symmetrize=False, remove_mean=False,
                              sparse_mode='sparse', sparse_tol=self.sparse_tol)
        self._test_moments_XY(self.X_100_sparsezero, self.Y_10_sparsezero, self.cols_10, symmetrize=False, remove_mean=False,
                              sparse_mode='sparse', sparse_tol=self.sparse_tol)
        self._test_moments_XY(self.X_100_sparsezero, self.Y_100_sparsezero, self.cols_100, symmetrize=False, remove_mean=False,
                              sparse_mode='sparse', sparse_tol=self.sparse_tol)
        # mean-free, sparse
        self._test_moments_XY(self.X_10_sparsezero, self.Y_10_sparsezero, self.cols_10, symmetrize=False, remove_mean=True,
                              sparse_mode='sparse', sparse_tol=self.sparse_tol)
        self._test_moments_XY(self.X_100_sparsezero, self.Y_10_sparsezero, self.cols_10, symmetrize=False, remove_mean=True,
                              sparse_mode='sparse', sparse_tol=self.sparse_tol)
        self._test_moments_XY(self.X_100_sparsezero, self.Y_100_sparsezero, self.cols_100, symmetrize=False, remove_mean=True,
                              sparse_mode='sparse', sparse_tol=self.sparse_tol)

    def test_moments_XY_weighted_sparsezero(self):
        # weighted test, sparse
        self._test_moments_XY(self.X_10_sparsezero, self.Y_10_sparsezero, self.cols_10, symmetrize=False, remove_mean=False,
                              sparse_mode='sparse', sparse_tol=self.sparse_tol, weights=self.weights)
        self._test_moments_XY(self.X_100_sparsezero, self.Y_100_sparsezero, self.cols_100, symmetrize=False, remove_mean=False,
                              sparse_mode='sparse', sparse_tol=self.sparse_tol, weights=self.weights)
        # weighted test, mean-free, sparse
        self._test_moments_XY(self.X_10_sparsezero, self.Y_10_sparsezero, self.cols_10, symmetrize=False, remove_mean=True,
                              sparse_mode='sparse', sparse_tol=self.sparse_tol, weights=self.weights)
        self._test_moments_XY(self.X_100_sparsezero, self.Y_100_sparsezero, self.cols_100, symmetrize=False, remove_mean=True,
                              sparse_mode='sparse', sparse_tol=self.sparse_tol, weights=self.weights)

    def test_moments_XY_sym_sparsezero(self):
        # simple test, sparse, symmetric
        self._test_moments_XY(self.X_10_sparsezero, self.Y_10_sparsezero, self.cols_10, symmetrize=True, remove_mean=False,
                              sparse_mode='sparse', sparse_tol=self.sparse_tol)
        self._test_moments_XY(self.X_100_sparsezero, self.Y_100_sparsezero, self.cols_100, symmetrize=True, remove_mean=False,
                              sparse_mode='sparse', sparse_tol=self.sparse_tol)
        # mean-free, sparse, symmetric
        self._test_moments_XY(self.X_10_sparsezero, self.Y_10_sparsezero, self.cols_10, symmetrize=True, remove_mean=True,
                              sparse_mode='sparse', sparse_tol=self.sparse_tol)
        self._test_moments_XY(self.X_100_sparsezero, self.Y_100_sparsezero, self.cols_100, symmetrize=True, remove_mean=True,
                              sparse_mode='sparse', sparse_tol=self.sparse_tol)

    def test_moments_XY_weighted_sym_sparsezero(self):
        # simple test, sparse, symmetric
        self._test_moments_XY(self.X_10_sparsezero, self.Y_10_sparsezero, self.cols_10, symmetrize=True, remove_mean=False,
                              sparse_mode='sparse', sparse_tol=self.sparse_tol, weights=self.weights)
        self._test_moments_XY(self.X_100_sparsezero, self.Y_100_sparsezero, self.cols_100, symmetrize=True, remove_mean=False,
                              sparse_mode='sparse', sparse_tol=self.sparse_tol, weights=self.weights)
        # mean-free, sparse, symmetric
        self._test_moments_XY(self.X_10_sparsezero, self.Y_10_sparsezero, self.cols_10, symmetrize=True, remove_mean=True,
                              sparse_mode='sparse', sparse_tol=self.sparse_tol, weights=self.weights)
        self._test_moments_XY(self.X_100_sparsezero, self.Y_100_sparsezero, self.cols_100, symmetrize=True, remove_mean=True,
                              sparse_mode='sparse', sparse_tol=self.sparse_tol, weights=self.weights)

    def test_moments_XY_sparseconst(self):
        # simple test, sparse
        self._test_moments_XY(self.X_10_sparseconst, self.Y_10_sparseconst, self.cols_10, symmetrize=False, remove_mean=False,
                              sparse_mode='sparse', sparse_tol=self.sparse_tol)
        self._test_moments_XY(self.X_100_sparseconst, self.Y_10_sparseconst, self.cols_10, symmetrize=False, remove_mean=False,
                              sparse_mode='sparse', sparse_tol=self.sparse_tol)
        self._test_moments_XY(self.X_100_sparseconst, self.Y_100_sparseconst, self.cols_100, symmetrize=False, remove_mean=False,
                              sparse_mode='sparse', sparse_tol=self.sparse_tol)
        # mean-free, sparse
        self._test_moments_XY(self.X_10_sparseconst, self.Y_10_sparseconst, self.cols_10, symmetrize=False, remove_mean=True,
                              sparse_mode='sparse', sparse_tol=self.sparse_tol)
        self._test_moments_XY(self.X_100_sparseconst, self.Y_10_sparseconst, self.cols_10, symmetrize=False, remove_mean=True,
                              sparse_mode='sparse', sparse_tol=self.sparse_tol)
        self._test_moments_XY(self.X_100_sparseconst, self.Y_100_sparseconst, self.cols_100, symmetrize=False, remove_mean=True,
                              sparse_mode='sparse', sparse_tol=self.sparse_tol)

    def test_moments_XY_weighted_sparseconst(self):
        # weighted test, sparse
        self._test_moments_XY(self.X_10_sparseconst, self.Y_10_sparseconst, self.cols_10, symmetrize=False, remove_mean=False,
                              sparse_mode='sparse', sparse_tol=self.sparse_tol, weights=self.weights)
        self._test_moments_XY(self.X_100_sparseconst, self.Y_100_sparseconst, self.cols_100, symmetrize=False, remove_mean=False,
                              sparse_mode='sparse', sparse_tol=self.sparse_tol, weights=self.weights)
        # weighted test, mean-free, sparse
        self._test_moments_XY(self.X_10_sparseconst, self.Y_10_sparseconst, self.cols_10, symmetrize=False, remove_mean=True,
                              sparse_mode='sparse', sparse_tol=self.sparse_tol, weights=self.weights)
        self._test_moments_XY(self.X_100_sparseconst, self.Y_100_sparseconst, self.cols_100, symmetrize=False, remove_mean=True,
                              sparse_mode='sparse', sparse_tol=self.sparse_tol, weights=self.weights)

    def test_moments_XY_sym_sparseconst(self):
        # simple test, sparse, symmetric
        self._test_moments_XY(self.X_10_sparseconst, self.Y_10_sparseconst, self.cols_10, symmetrize=True, remove_mean=False,
                              sparse_mode='sparse', sparse_tol=self.sparse_tol)
        self._test_moments_XY(self.X_100_sparseconst, self.Y_100_sparseconst, self.cols_100, symmetrize=True, remove_mean=False,
                              sparse_mode='sparse', sparse_tol=self.sparse_tol)
        # mean-free, sparse, symmetric
        self._test_moments_XY(self.X_10_sparseconst, self.Y_10_sparseconst, self.cols_10, symmetrize=True, remove_mean=True,
                              sparse_mode='sparse', sparse_tol=self.sparse_tol)
        self._test_moments_XY(self.X_100_sparseconst, self.Y_100_sparseconst, self.cols_100, symmetrize=True, remove_mean=True,
                              sparse_mode='sparse', sparse_tol=self.sparse_tol)

    def test_moments_XY_weighted_sym_sparseconst(self):
        # simple test, sparse, symmetric
        self._test_moments_XY(self.X_10_sparseconst, self.Y_10_sparseconst, self.cols_10, symmetrize=True, remove_mean=False,
                              sparse_mode='sparse', sparse_tol=self.sparse_tol, weights=self.weights)
        self._test_moments_XY(self.X_100_sparseconst, self.Y_100_sparseconst, self.cols_100, symmetrize=True, remove_mean=False,
                              sparse_mode='sparse', sparse_tol=self.sparse_tol, weights=self.weights)
        # mean-free, sparse, symmetric
        self._test_moments_XY(self.X_10_sparseconst, self.Y_10_sparseconst, self.cols_10, symmetrize=True, remove_mean=True,
                              sparse_mode='sparse', sparse_tol=self.sparse_tol, weights=self.weights)
        self._test_moments_XY(self.X_100_sparseconst, self.Y_100_sparseconst, self.cols_100, symmetrize=True, remove_mean=True,
                              sparse_mode='sparse', sparse_tol=self.sparse_tol, weights=self.weights)

    def _test_moments_block(self, X, Y, column_selection, remove_mean=False, sparse_mode='auto', sparse_tol=0.0):
        w1, s, C = moments.moments_block(X, Y, remove_mean=remove_mean, modify_data=False,
                                         sparse_mode=sparse_mode, sparse_tol=sparse_tol)
        # reference
        w = X.shape[0]
        s_X_ref = X.sum(axis=0)
        s_Y_ref = Y.sum(axis=0)
        if remove_mean:
            X = X - s_X_ref/float(w)
            Y = Y - s_Y_ref/float(w)
        C_XX_ref = np.dot(X.T, X)
        C_XY_ref = np.dot(X.T, Y)
        C_YX_ref = np.dot(Y.T, X)
        C_YY_ref = np.dot(Y.T, Y)
        # test
        assert np.allclose(w1, w)
        assert np.allclose(s[0], s_X_ref)
        assert np.allclose(s[1], s_Y_ref)
        assert np.allclose(C[0][0], C_XX_ref)
        assert np.allclose(C[0][1], C_XY_ref)
        assert np.allclose(C[1][0], C_YX_ref)
        assert np.allclose(C[1][1], C_YY_ref)
        # column subsets
        w1, s, C = moments.moments_block(X, Y, remove_mean=remove_mean, modify_data=False,
                                         sparse_mode=sparse_mode, sparse_tol=sparse_tol, column_selection=column_selection)
        assert np.allclose(C[0][0], C_XX_ref[:, column_selection])
        assert np.allclose(C[0][1], C_XY_ref[:, column_selection])
        assert np.allclose(C[1][0], C_YX_ref[:, column_selection])
        assert np.allclose(C[1][1], C_YY_ref[:, column_selection])
        # diagonal only
        if sparse_mode is not 'sparse' and X.shape[1] == Y.shape[1]:
            w1, s, C = moments.moments_block(X, Y, remove_mean=remove_mean, modify_data=False,
                                             sparse_mode=sparse_mode, sparse_tol=sparse_tol, diag_only=True)
            assert np.allclose(C[0][0], np.diag(C_XX_ref))
            assert np.allclose(C[0][1], np.diag(C_XY_ref))
            assert np.allclose(C[1][0], np.diag(C_YX_ref))
            assert np.allclose(C[1][1], np.diag(C_YY_ref))

    def test_moments_block(self):
        # simple test, dense
        self._test_moments_block(self.X_10, self.Y_10, self.cols_10, remove_mean=False, sparse_mode='dense')
        self._test_moments_block(self.X_100, self.Y_100, self.cols_100, remove_mean=False, sparse_mode='dense')
        # mean-free, dense
        self._test_moments_block(self.X_10, self.Y_10, self.cols_10, remove_mean=True, sparse_mode='dense')
        self._test_moments_block(self.X_100, self.Y_100, self.cols_100, remove_mean=True, sparse_mode='dense')

    def test_moments_block_sparsezero(self):
        # simple test, sparse
        self._test_moments_block(self.X_10_sparsezero, self.Y_10_sparsezero, self.cols_10, remove_mean=False,
                                 sparse_mode='sparse', sparse_tol=self.sparse_tol)
        self._test_moments_block(self.X_100_sparsezero, self.Y_100_sparsezero, self.cols_100, remove_mean=False,
                                 sparse_mode='sparse', sparse_tol=self.sparse_tol)
        # mean-free, sparse
        self._test_moments_block(self.X_10_sparsezero, self.Y_10_sparsezero, self.cols_10, remove_mean=True,
                                 sparse_mode='sparse', sparse_tol=self.sparse_tol)
        self._test_moments_block(self.X_100_sparsezero, self.Y_100_sparsezero, self.cols_100, remove_mean=True,
                                 sparse_mode='sparse', sparse_tol=self.sparse_tol)

    def test_moments_block_sparseconst(self):
        # simple test, sparse
        self._test_moments_block(self.X_10_sparseconst, self.Y_10_sparseconst, self.cols_10, remove_mean=False,
                                 sparse_mode='sparse', sparse_tol=self.sparse_tol)
        self._test_moments_block(self.X_100_sparseconst, self.Y_100_sparseconst, self.cols_100, remove_mean=False,
                                 sparse_mode='sparse', sparse_tol=self.sparse_tol)
        # mean-free, sparse
        self._test_moments_block(self.X_10_sparseconst, self.Y_10_sparseconst, self.cols_10, remove_mean=True,
                                 sparse_mode='sparse', sparse_tol=self.sparse_tol)
        self._test_moments_block(self.X_100_sparseconst, self.Y_100_sparseconst, self.cols_100, remove_mean=True,
                                 sparse_mode='sparse', sparse_tol=self.sparse_tol)


if __name__ == "__main__":
    unittest.main()




# This file is part of PyEMMA.
#
# Copyright (c) 2015, 2014 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from pyemma._base.model import SampledModel
from pyemma.msm.models.msm import MSM

__author__ = 'noe'


class SampledMSM(MSM, SampledModel):
    r""" Sampled Markov state model """

    __serialize_version = 0

    # TODO: maybe rename to parametrize in order to avoid confusion with set_params that has a different behavior?
    def set_model_params(self, samples=None, conf=0.95,
                         P=None, pi=None, reversible=None, dt_model='1 step', neig=None):
        """

        Parameters
        ----------
        samples : list of MSM objects
            sampled MSMs
        conf : float, optional, default=0.68
            Confidence interval. By default one-sigma (68.3%) is used. Use 95.4% for two sigma or 99.7% for three sigma.

        """
        # set model parameters of superclass
        SampledModel.set_model_params(self, samples=samples, conf=conf)
        MSM.set_model_params(self, P=P, pi=pi, reversible=reversible, dt_model=dt_model, neig=neig)




# This file is part of PyEMMA.
#
# Copyright (c) 2015, 2014 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
from unittest import TestCase
import numpy as np
from pyemma.coordinates.api import cluster_mini_batch_kmeans


class TestMiniBatchKmeans(TestCase):
    def test_3gaussian_1d_singletraj(self):
        # generate 1D data from three gaussians
        X = [np.random.randn(200) - 2.0,
             np.random.randn(300),
             np.random.randn(400) + 2.0]
        X = np.hstack(X)
        kmeans = cluster_mini_batch_kmeans(X, batch_size=0.5, k=100, max_iter=10000)
        cc = kmeans.clustercenters
        assert (np.any(cc < 1.0))
        assert (np.any((cc > -1.0) * (cc < 1.0)))
        assert (np.any(cc > -1.0))

    def test_3gaussian_2d_multitraj(self):
        # generate 1D data from three gaussians
        X1 = np.zeros((200, 2))
        X1[:, 0] = np.random.randn(200) - 2.0
        X2 = np.zeros((300, 2))
        X2[:, 0] = np.random.randn(300)
        X3 = np.zeros((400, 2))
        X3[:, 0] = np.random.randn(400) + 2.0
        X = [X1, X2, X3]
        kmeans = cluster_mini_batch_kmeans(X, batch_size=0.5, k=100, max_iter=10000)
        cc = kmeans.clustercenters
        assert (np.any(cc < 1.0))
        assert (np.any((cc > -1.0) * (cc < 1.0)))
        assert (np.any(cc > -1.0))


class TestMiniBatchKmeansResume(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        from pyemma.util.contexts import numpy_random_seed
        with numpy_random_seed(32):
            # three gaussians
            X = [np.random.randn(1000)-2.0,
                 np.random.randn(1000),
                 np.random.randn(1000)+2.0]
            cls.X = np.hstack(X)

    def test_resume(self):
        """ check that we can continue with the iteration by passing centers"""
        # centers are far off
        initial_centers = np.array([[1, 2, 3]]).T
        cl = cluster_mini_batch_kmeans(self.X, clustercenters=initial_centers,
                                       max_iter=1, k=3)

        resume_centers = cl.clustercenters
        cl.estimate(self.X, clustercenters=resume_centers, max_iter=50)
        new_centers = cl.clustercenters

        true = np.array([[-2, 0, 2]]).T
        d0 = true - resume_centers
        d1 = true - new_centers

        diff = np.linalg.norm(d0)
        diff_next = np.linalg.norm(d1)

        self.assertLess(diff_next, diff, 'resume_centers=%s, new_centers=%s' % (resume_centers, new_centers))

if __name__ == '__main__':
    unittest.main()




# This file is part of PyEMMA.
#
# Copyright (c) 2015, 2014 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


__author__ = 'noe'

import numpy as np
from pyemma.msm import markov_model

class DoubleWell_Discrete_Data(object):
    """ MCMC process in a symmetric double well potential, spatially discretized to 100 bins

    """

    def __init__(self):
        from pkg_resources import resource_filename
        filename = resource_filename('pyemma.datasets', 'double_well_discrete.npz')
        datafile = np.load(filename)
        self._dtraj_T100K_dt10 = datafile['dtraj']
        self._P = datafile['P']
        self._msm = markov_model(self._P)

    @property
    def dtraj_T100K_dt10(self):
        """ 100K frames trajectory at timestep 10, 100 microstates (not all are populated). """
        return self._dtraj_T100K_dt10

    @property
    def dtraj_T100K_dt10_n2good(self):
        """ 100K frames trajectory at timestep 10, good 2-state discretization (at transition state). """
        return self.dtraj_T100K_dt10_n([50])

    @property
    def dtraj_T100K_dt10_n2bad(self):
        """ 100K frames trajectory at timestep 10, bad 2-state discretization (off transition state). """
        return self.dtraj_T100K_dt10_n([40])

    def dtraj_T100K_dt10_n2(self, divide):
        """ 100K frames trajectory at timestep 10, arbitrary 2-state discretization. """
        return self.dtraj_T100K_dt10_n([divide])

    @property
    def dtraj_T100K_dt10_n6good(self):
        """ 100K frames trajectory at timestep 10, good 6-state discretization. """
        return self.dtraj_T100K_dt10_n([40, 45, 50, 55, 60])

    def dtraj_T100K_dt10_n(self, divides):
        """ 100K frames trajectory at timestep 10, arbitrary n-state discretization. """
        disc = np.zeros(100, dtype=int)
        divides = np.concatenate([divides, [100]])
        for i in range(len(divides)-1):
            disc[divides[i]:divides[i+1]] = i+1
        return disc[self.dtraj_T100K_dt10]

    @property
    def transition_matrix(self):
        """ Exact transition matrix used to generate the data """
        return self._P

    @property
    def msm(self):
        """ Returns an MSM object with the exact transition matrix """
        return self._msm

    def generate_traj(self, N, start=None, stop=None, dt=1):
        """ Generates a random trajectory of length N with time step dt """
        from msmtools.generation import generate_traj
        return generate_traj(self._P, N, start=start, stop=stop, dt=dt)

    def generate_trajs(self, M, N, start=None, stop=None, dt=1):
        """ Generates M random trajectories of length N each with time step dt """
        from msmtools.generation import generate_trajs
        return generate_trajs(self._P, M, N, start=start, stop=stop, dt=dt)




# This file is part of PyEMMA.
#
# Copyright (c) 2015, 2014 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
Created on May 26, 2014

@author: marscher
'''


class SpectralWarning(RuntimeWarning):
    pass


class ImaginaryEigenValueWarning(SpectralWarning):
    pass


class PrecisionWarning(RuntimeWarning):
    r"""
    This warning indicates that some operation in your code leads
    to a conversion of datatypes, which involves a loss/gain in
    precision.

    """
    pass


class NotConvergedWarning(RuntimeWarning):
    r"""
    This warning indicates that some iterative procdure has not
    converged or reached the maximum number of iterations implemented
    as a safe guard to prevent arbitrary many iterations in loops with
    a conditional termination criterion.

    """
    pass


class EfficiencyWarning(UserWarning):
    r"""Some operation or input data leads to a lack of efficiency"""
    pass


class ParserWarning(UserWarning):
    """ Some user defined variable could not be parsed and is ignored/replaced. """
    pass


class ConfigDirectoryException(Exception):
    """ Some operation with PyEMMAs configuration directory went wrong. """
    pass


class PyEMMA_DeprecationWarning(UserWarning):
    """You are using a feature, which will be removed in a future release. You have been warned!"""
    pass


class PyEMMA_UserWarning(UserWarning):
    """ general user warning."""
    pass



"""

Data Types
----------
The standard data type for covariance computations is
float64, because the double precision (but not single precision) is
usually sufficient to compute the long sums involved in covariance
matrix computations. Integer types are avoided even if the data is integer,
because the BLAS matrix multiplication is very fast with floats, but very
slow with integers. If X is of boolean type (0/1), the standard data type
is float32, because this will be sufficient to represent numbers up to 2^23
without rounding error, which is usually sufficient sufficient as the
largest element in np.dot(X.T, X) can then be T, the number of data points.

Efficient Use
-------------
In order to get speedup with boolean input, remove_mean=False is required.
Note that you can still do TICA that way.

Sparsification
--------------
We aim at computing covariance matrices. For large (T x N) data matrices X, Y,
the bottleneck of this operation is computing the matrix product np.dot(X.T, X),
or np.dot(X.T, Y), with algorithmic complexity O(N^2 T). If X, Y have zero or
constant columns, we can reduce N and thus reduce the algorithmic complexity.

However, the BLAS matrix product used by np.dot() is highly Cache optimized -
the data is accessed in a way that most operations are done in cache, making the
calculation extremely efficient. Thus, even if X, Y have zero or constant columns,
it does not always pay off to interfere with this operation - one one hand by
spending compute time to determine the sparsity of the matrices, one the other
hand by using slicing operations that reduce the algorithmic complexity, but may
destroy the order of the data and thus produce more cache failures.

In order to make an informed decision, we have compared the runtime of the following
operations using matrices of various different sizes (T x N) and different degrees
of sparsity. (using an Intel Core i7 with OS/X 10.10.1):

    1. Compute np.dot(X.T, X)
    2. Compute np.dot(X[:, sel].T, X[:, sel]) where sel selects the nonzero columns
    3. Make a copy X0 = X[:, sel].copy() and then compute np.dot(X0.T, X0)

It may seem that step 3 is not a good idea because we make the extra effort of
copying the matrix. However, the new copy will have data ordered sequentially in
memory, and therefore better prepared for the algorithmically more expensive but
cache-optimized matrix product.

We have empirically found that:

    * Making a copy before running np.dot (option 3) is in most cases better than
      using the dot product on sliced arrays (option 2). Exceptions are when the
      data is extremely sparse, such that only a few columns are selected.
    * Copying and subselecting columns (option 3) is only faster than the full
      dot product (option 1), if 50% or less columns are selected. This observation
      is roughly independent of N.
    * The observations above are valid for  matrices (T x N) that are sufficiently
      large. We assume that "sufficiently large" means that they don't fully fit
      in the cache. For small matrices, the trends are less clear and different
      rules may apply.

In order to optimize covariance calculation for large matrices, we therefore
take the following actions:

    1. Given matrix size of X (and Y), determine the minimum number of columns
       that need to be constant in order to use sparse computation.
    2. Efficiently determine sparsity of X (and Y). Give up as soon as the
       number of constant column candidates drops below the minimum number, to
       avoid wasting time on the decision.
    3. Subselect the desired columns and copy the data to a new array X0 (Y0).
    4. Run operation on the new array X0 (Y0), including in-place substraction
       of the mean if needed.

"""

__author__ = 'noe'

import math
import numbers
import numpy as np
from .covar_c import covartools


def _is_zero(x):
    """ Returns True if x is numerically 0 or an array with 0's. """
    if x is None:
        return True
    if isinstance(x, numbers.Number):
        return x == 0.0
    if isinstance(x, np.ndarray):
        return np.all(x == 0)
    return False


def _sparsify(X, remove_mean=False, modify_data=False, sparse_mode='auto', sparse_tol=0.0):
    """ Determines the sparsity of X and returns a selected sub-matrix

    Only conducts sparsification if the number of constant columns is at least
    max(a N - b, min_const_col_number),

    Parameters
    ----------
    X : ndarray
        data matrix
    remove_mean : bool
        True: remove column mean from the data, False: don't remove mean.
    modify_data : bool
        If remove_mean=True, the mean will be removed in the data matrix X,
        without creating an independent copy. This option is faster but might
        lead to surprises because your input array is changed.
    sparse_mode : str
        one of:
            * 'dense' : always use dense mode
            * 'sparse' : always use sparse mode if possible
            * 'auto' : automatic

    Returns
    -------
    X0 : ndarray (view of X)
        Either X itself (if not sufficiently sparse), or a sliced view of X,
        containing only the variable columns
    mask : ndarray(N, dtype=bool) or None
        Bool selection array that indicates which columns of X were selected for
        X0, i.e. X0 = X[:, mask]. mask is None if no sparse selection was made.
    xconst : ndarray(N)
        Constant column values that are outside the sparse selection, i.e.
        X[i, ~mask] = xconst for any row i. xconst=0 if no sparse selection was made.

    """
    if sparse_mode.lower() == 'sparse':
        min_const_col_number = 0  # enforce sparsity. A single constant column will lead to sparse treatment
    elif sparse_mode.lower() == 'dense':
        min_const_col_number = X.shape[1] + 1  # never use sparsity
    else:
        if remove_mean and not modify_data:  # in this case we have to copy the data anyway, and can be permissive
            min_const_col_number = max(0.1 * X.shape[1], 50)
        else:
            # This is a rough heuristic to choose a minimum column number for which sparsity may pay off.
            # This heuristic is good for large number of samples, i.e. it may be inadequate for small matrices X.
            if X.shape[1] < 250:
                min_const_col_number = X.shape[1] - 0.25 * X.shape[1]
            elif X.shape[1] < 1000:
                min_const_col_number = X.shape[1] - (0.5 * X.shape[1] - 100)
            else:
                min_const_col_number = X.shape[1] - (0.8 * X.shape[1] - 400)
        # ensure we have an integer again.
        min_const_col_number = int(min_const_col_number)

    if X.shape[1] > min_const_col_number:
        mask = covartools.variable_cols(X, tol=sparse_tol, min_constant=min_const_col_number)  # bool vector
        nconst = len(np.where(~mask)[0])
        if nconst > min_const_col_number:
            xconst = X[0, ~mask]
            X = X[:, mask]  # sparsify
        else:
            xconst = None
            mask = None
    else:
        xconst = None
        mask = None

    return X, mask, xconst  # None, 0 if not sparse


def _sparsify_pair(X, Y, remove_mean=False, modify_data=False, symmetrize=False, sparse_mode='auto', sparse_tol=0.0):
    """
    """
    T = X.shape[0]
    N = math.sqrt(X.shape[1] * Y.shape[1])
    # check each data set separately for sparsity.
    X0, mask_X, xconst = _sparsify(X, sparse_mode=sparse_mode, sparse_tol=sparse_tol)
    Y0, mask_Y, yconst = _sparsify(Y, sparse_mode=sparse_mode, sparse_tol=sparse_tol)
    # if we have nonzero constant columns and the number of samples is too small, do not treat as
    # sparse, because then the const-specialized dot product function doesn't pay off.
    is_const = not (_is_zero(xconst) and _is_zero(yconst))
    if is_const and (symmetrize or not remove_mean) and 10*T < N:
        return X, None, None, Y, None, None
    else:
        return X0, mask_X, xconst, Y0, mask_Y, yconst


def _copy_convert(X, const=None, remove_mean=False, copy=True):
    r""" Makes a copy or converts the data type if needed

    Copies the data and converts the data type if unsuitable for covariance
    calculation. The standard data type for covariance computations is
    float64, because the double precision (but not single precision) is
    usually sufficient to compute the long sums involved in covariance
    matrix computations. Integer types are avoided even if the data is integer,
    because the BLAS matrix multiplication is very fast with floats, but very
    slow with integers. If X is of boolean type (0/1), the standard data type
    is float32, because this will be sufficient to represent numbers up to 2^23
    without rounding error, which is usually sufficient sufficient as the
    largest element in np.dot(X.T, X) can then be T, the number of data points.

    Parameters
    ----------
    remove_mean : bool
        If True, will enforce float64 even if the input is boolean
    copy : bool
        If True, enforces a copy even if the data type doesn't require it.

    Return
    ------
    X : ndarray
        copy or reference to X if no copy was needed.
    const : ndarray or None
        copy or reference to const if no copy was needed.

    """
    # determine type
    dtype = np.float64  # default: convert to float64 in order to avoid cancellation errors
    if X.dtype.kind == 'b' and X.shape[0] < 2**23 and not remove_mean:
        dtype = np.float32  # convert to float32 if we can represent all numbers
    # copy/convert if needed
    if X.dtype not in (np.float64, dtype):  # leave as float64 (conversion is expensive), otherwise convert to dtype
        X = X.astype(dtype, order='C')
        if const is not None:
            const = const.astype(dtype, order='C')
    elif copy:
        X = X.copy(order='C')
        if const is not None:
            const = const.copy(order='C')

    return X, const


def _sum_sparse(xsum, mask_X, xconst, T):
    s = np.zeros(len(mask_X))
    s[mask_X] = xsum
    s[~mask_X] = T * xconst
    return s


def _sum(X, xmask=None, xconst=None, Y=None, ymask=None, yconst=None, symmetric=False, remove_mean=False,
         weights=None):
    r""" Computes the column sums and centered column sums.

    If symmetric = False, the sums will be determined as
    .. math:
        sx &=& \frac{1}{2} \sum_t x_t
        sy &=& \frac{1}{2} \sum_t y_t

    If symmetric, the sums will be determined as

    .. math:
        sx = sy = \frac{1}{2T} \sum_t x_t + y_t

    Returns
    -------
    w : float
        statistical weight of sx, sy
    sx : ndarray
        effective row sum of X (including symmetrization if requested)
    sx_raw_centered : ndarray
        centered raw row sum of X

    optional returns (only if Y is given):

    sy : ndarray
        effective row sum of X (including symmetrization if requested)
    sy_raw_centered : ndarray
        centered raw row sum of Y

    """
    T = X.shape[0]
    # Check if weights are given:
    if weights is not None:
        X = weights[:, None] * X
        if Y is not None:
            Y = weights[:, None] * Y
    # compute raw sums on variable data
    sx_raw = X.sum(axis=0)  # this is the mean before subtracting it.
    sy_raw = 0
    if Y is not None:
        sy_raw = Y.sum(axis=0)

    # expand raw sums to full data
    if xmask is not None:
        if weights is not None:
            sx_raw = _sum_sparse(sx_raw, xmask, xconst, weights.sum())
        else:
            sx_raw = _sum_sparse(sx_raw, xmask, xconst, T)
    if ymask is not None:
        if weights is not None:
            sy_raw = _sum_sparse(sy_raw, ymask, yconst, weights.sum())
        else:
            sy_raw = _sum_sparse(sy_raw, ymask, yconst, T)

    # compute effective sums and centered sums
    if Y is not None and symmetric:
        sx = sx_raw + sy_raw
        sy = sx
        if weights is not None:
            w = 2*np.sum(weights)
        else:
            w = 2 * T
    else:
        sx = sx_raw
        sy = sy_raw
        if weights is not None:
            w = np.sum(weights)
        else:
            w = T

    sx_raw_centered = sx_raw.copy()
    if Y is not None:
        sy_raw_centered = sy_raw.copy()

    # center mean.
    if remove_mean:
        if Y is not None and symmetric:
            sx_raw_centered -= 0.5 * sx
            sy_raw_centered -= 0.5 * sy
        else:
            sx_raw_centered = np.zeros(sx.size)
            if Y is not None:
                sy_raw_centered = np.zeros(sy.size)

    # return
    if Y is not None:
        return w, sx, sx_raw_centered, sy, sy_raw_centered
    else:
        return w, sx, sx_raw_centered


def _center(X, w, s, mask=None, const=None, inplace=True):
    """ Centers the data.

    Parameters
    ----------
    w : float
        statistical weight of s
    inplace : bool
        center in place

    Returns
    -------
    sx : ndarray
        uncentered row sum of X
    sx_centered : ndarray
        row sum of X after centering

    optional returns (only if Y is given):

    sy_raw : ndarray
        uncentered row sum of Y
    sy_centered : ndarray
        row sum of Y after centering

    """
    xmean = s / float(w)
    if mask is None:
        X = np.subtract(X, xmean, out=X if inplace else None)
    else:
        X = np.subtract(X, xmean[mask], out=X if inplace else None)
        const = np.subtract(const, xmean[~mask], const if inplace else None)

    return X, const


def _filter_variable_indices(mask, column_selection):
    """ Returns column indices restricted to the variable columns as determined by the given mask.

    Parameters
    ----------
    mask : ndarray(N, dtype=bool)
        Array indicating the variable columns.
    column_selection : ndarray(k, dtype=int)
        Column indices to be filtered and mapped.

    Returns
    -------
    ix : ndarray(l, dtype=int)
        Column indices restricted to the variable columns, mapped to the correct index range.

    """
    a = np.where(mask)[0]
    b = column_selection[np.in1d(column_selection, a)]
    return np.searchsorted(a, b)


# ====================================================================================
# SECOND MOMENT MATRICES / COVARIANCES
# ====================================================================================

def _M2_dense(X, Y, weights=None, diag_only=False):
    """ 2nd moment matrix using dense matrix computations.

    This function is encapsulated such that we can make easy modifications of the basic algorithms

    """
    if weights is not None:
        if diag_only:
            return np.sum(weights[:, None] * X * Y, axis=0)
        else:
            return np.dot((weights[:, None] * X).T, Y)
    else:
        if diag_only:
            return np.sum(X * Y, axis=0)
        else:
            return np.dot(X.T, Y)


def _M2_const(Xvar, mask_X, xvarsum, xconst, Yvar, mask_Y, yvarsum, yconst, weights=None):
    r""" Computes the unnormalized covariance matrix between X and Y, exploiting constant input columns

    Computes the unnormalized covariance matrix :math:`C = X^\top Y`
    (for symmetric=False) or :math:`C = \frac{1}{2} (X^\top Y + Y^\top X)`
    (for symmetric=True). Suppose the data matrices can be column-permuted
    to have the form

    .. math:
        X &=& (X_{\mathrm{var}}, X_{\mathrm{const}})
        Y &=& (Y_{\mathrm{var}}, Y_{\mathrm{const}})

    with rows:

    .. math:
        x_t &=& (x_{\mathrm{var},t}, x_{\mathrm{const}})
        y_t &=& (y_{\mathrm{var},t}, y_{\mathrm{const}})

    where :math:`x_{\mathrm{const}},\:y_{\mathrm{const}}` are constant vectors.
    The resulting matrix has the general form:

    .. math:
        C &=& [X_{\mathrm{var}}^\top Y_{\mathrm{var}}  x_{sum} y_{\mathrm{const}}^\top ]
          & & [x_{\mathrm{const}}^\top y_{sum}^\top    x_{sum} x_{sum}^\top            ]

    where :math:`x_{sum} = \sum_t x_{\mathrm{var},t}` and
    :math:`y_{sum} = \sum_t y_{\mathrm{var},t}`.

    Parameters
    ----------
    Xvar : ndarray (T, m)
        Part of the data matrix X with :math:`m \le M` variable columns.
    mask_X : ndarray (M)
        Boolean array of size M of the full columns. False for constant column,
        True for variable column in X.
    xvarsum : ndarray (m)
        Column sum of variable part of data matrix X
    xconst : ndarray (M-m)
        Values of the constant part of data matrix X
    Yvar : ndarray (T, n)
        Part of the data matrix Y with :math:`n \le N` variable columns.
    mask_Y : ndarray (N)
        Boolean array of size N of the full columns. False for constant column,
        True for variable column in Y.
    yvarsum : ndarray (n)
        Column sum of variable part of data matrix Y
    yconst : ndarray (N-n)
        Values of the constant part of data matrix Y
    weights : None or ndarray (N)
        weights for all time steps.

    Returns
    -------
    C : ndarray (M, N)
        Unnormalized covariance matrix.

    """
    C = np.zeros((len(mask_X), len(mask_Y)))
    # Block 11
    C[np.ix_(mask_X, mask_Y)] = _M2_dense(Xvar, Yvar, weights=weights)
    # other blocks
    xsum_is_0 = _is_zero(xvarsum)
    ysum_is_0 = _is_zero(yvarsum)
    xconst_is_0 = _is_zero(xconst)
    yconst_is_0 = _is_zero(yconst)
    # TODO: maybe we don't need the checking here, if we do the decision in the higher-level function M2
    # TODO: if not zero, we could still exploit the zeros in const and compute (and write!) this outer product
    # TODO: only to a sub-matrix
    # Block 12 and 21
    if weights is not None:
        wsum = np.sum(weights)
        xvarsum = np.sum(weights[:, None] * Xvar, axis=0)
        yvarsum = np.sum(weights[:, None] * Yvar, axis=0)
    else:
        wsum = Xvar.shape[0]
    if not (xsum_is_0 or yconst_is_0) or not (ysum_is_0 or xconst_is_0):
        C[np.ix_(mask_X, ~mask_Y)] = np.outer(xvarsum, yconst)
        C[np.ix_(~mask_X, mask_Y)] = np.outer(xconst, yvarsum)
    # Block 22
    if not (xconst_is_0 or yconst_is_0):
        C[np.ix_(~mask_X, ~mask_Y)] = np.outer(wsum*xconst, yconst)
    return C


def _M2_sparse(Xvar, mask_X, Yvar, mask_Y, weights=None):
    """ 2nd moment matrix exploiting zero input columns """
    C = np.zeros((len(mask_X), len(mask_Y)))
    C[np.ix_(mask_X, mask_Y)] = _M2_dense(Xvar, Yvar, weights=weights)
    return C


def _M2_sparse_sym(Xvar, mask_X, Yvar, mask_Y, weights=None, column_selection=None):
    """ 2nd self-symmetric moment matrix exploiting zero input columns

    Computes X'X + Y'Y and X'Y + Y'X

    """
    assert len(mask_X) == len(mask_Y), 'X and Y need to have equal sizes for symmetrization'

    if column_selection is None:
        mask_Xk = mask_X
        mask_Yk = mask_Y
        Xvark = Xvar
        Yvark = Yvar
    else:
        mask_Xk = mask_X[column_selection]
        mask_Yk = mask_Y[column_selection]
        Xvark = Xvar[:, _filter_variable_indices(mask_X, column_selection)]
        Yvark = Yvar[:, _filter_variable_indices(mask_Y, column_selection)]

    Cxxyy = np.zeros((len(mask_X), len(mask_Yk)))
    Cxxyy[np.ix_(mask_X, mask_Xk)] = _M2_dense(Xvar, Xvark, weights=weights)
    Cxxyy[np.ix_(mask_Y, mask_Yk)] += _M2_dense(Yvar, Yvark, weights=weights)

    Cxyyx = np.zeros((len(mask_X), len(mask_Yk)))
    Cxy = _M2_dense(Xvar, Yvark, weights=weights)
    Cyx = _M2_dense(Yvar, Xvark, weights=weights)
    Cxyyx[np.ix_(mask_X, mask_Yk)] = Cxy
    Cxyyx[np.ix_(mask_Y, mask_Xk)] += Cyx

    return Cxxyy, Cxyyx


def _M2(Xvar, Yvar, mask_X=None, mask_Y=None, xsum=0, xconst=0, ysum=0, yconst=0, weights=None, diag_only=False):
    """ direct (nonsymmetric) second moment matrix. Decide if we need dense, sparse, const"""
    if mask_X is None and mask_Y is None:
        return _M2_dense(Xvar, Yvar, weights=weights, diag_only=diag_only)
    else:
        # Check if one of the masks is not None, modify it and also adjust the constant columns:
        if mask_X is None:
            mask_X = np.ones(Xvar.shape[1], dtype=np.bool)
            xconst = np.ones(0, dtype=float)
        if mask_Y is None:
            mask_Y = np.ones(Yvar.shape[1], dtype=np.bool)
            yconst = np.ones(0, dtype=float)
    if _is_zero(xsum) and _is_zero(ysum) or _is_zero(xconst) and _is_zero(yconst):
        return _M2_sparse(Xvar, mask_X, Yvar, mask_Y, weights=weights)
    else:
        return _M2_const(Xvar, mask_X, xsum[mask_X], xconst, Yvar, mask_Y, ysum[mask_Y], yconst, weights=weights)


def _M2_symmetric(Xvar, Yvar, mask_X=None, mask_Y=None, xsum=0, xconst=0, ysum=0, yconst=0, weights=None,
                  column_selection=None, diag_only=False):
    """ symmetric second moment matrices. Decide if we need dense, sparse, const"""
    if mask_X is None and mask_Y is None:
        if column_selection is None:
            Xvark = Xvar
            Yvark = Yvar
        else:
            Xvark = Xvar[:, column_selection]
            Yvark = Yvar[:, column_selection]
        Cxxyy = _M2_dense(Xvar, Xvark, weights=weights, diag_only=diag_only) \
                + _M2_dense(Yvar, Yvark, weights=weights, diag_only=diag_only)
        Cxy = _M2_dense(Xvar, Yvark, weights=weights, diag_only=diag_only)
        Cyx = _M2_dense(Yvar, Xvark, weights=weights, diag_only=diag_only)
        Cxyyx = Cxy + Cyx
    else:
        # Check if one of the masks is not None, modify it and also adjust the constant columns:
        if mask_X is None:
            mask_X = np.ones(Xvar.shape[1], dtype=np.bool)
            xconst = np.ones(0, dtype=float)
        if mask_Y is None:
            mask_Y = np.ones(Yvar.shape[1], dtype=np.bool)
            yconst = np.ones(0, dtype=float)
        if _is_zero(xsum) and _is_zero(ysum) or _is_zero(xconst) and _is_zero(yconst):
            Cxxyy, Cxyyx = _M2_sparse_sym(Xvar, mask_X, Yvar, mask_Y, weights=weights, column_selection=column_selection)
        else:
            xvarsum = xsum[mask_X]  # to variable part
            yvarsum = ysum[mask_Y]  # to variable part
            if column_selection is None:
                Xvark = Xvar
                mask_Xk = mask_X
                xkvarsum = xvarsum
                xkconst = xconst
                Yvark = Yvar
                mask_Yk = mask_Y
                ykvarsum = yvarsum
                ykconst = yconst
            else:
                Xvark = Xvar[:, _filter_variable_indices(mask_X, column_selection)]
                mask_Xk = mask_X[column_selection]
                xksum = xsum[column_selection]
                xkvarsum = xksum[mask_Xk]
                xkconst = xconst[_filter_variable_indices(~mask_X, column_selection)]
                Yvark = Yvar[:, _filter_variable_indices(mask_Y, column_selection)]
                mask_Yk = mask_Y[column_selection]
                yksum = ysum[column_selection]
                ykvarsum = yksum[mask_Yk]
                ykconst = yconst[_filter_variable_indices(~mask_Y, column_selection)]
            Cxxyy = _M2_const(Xvar, mask_X, xvarsum, xconst, Xvark, mask_Xk, xkvarsum, xkconst, weights=weights) \
                    + _M2_const(Yvar, mask_Y, yvarsum, yconst, Yvark, mask_Yk, ykvarsum, ykconst, weights=weights)
            Cxy = _M2_const(Xvar, mask_X, xvarsum, xconst, Yvark, mask_Yk, ykvarsum, ykconst, weights=weights)
            Cyx = _M2_const(Yvar, mask_Y, yvarsum, yconst, Xvark, mask_Xk, xkvarsum, xkconst, weights=weights)
            Cxyyx = Cxy + Cyx
    return Cxxyy, Cxyyx


# =================================================
# USER API
# =================================================


def moments_XX(X, remove_mean=False, modify_data=False, weights=None, sparse_mode='auto', sparse_tol=0.0,
               column_selection=None, diag_only=False):
    r""" Computes the first two unnormalized moments of X

    Computes :math:`s = \sum_t x_t` and :math:`C = X^\top X` while exploiting
    zero or constant columns in the data matrix.

    Parameters
    ----------
    X : ndarray (T, M)
        Data matrix
    remove_mean : bool
        True: remove column mean from the data, False: don't remove mean.
    modify_data : bool
        If remove_mean=True, the mean will be removed in the data matrix X,
        without creating an independent copy. This option is faster but might
        lead to surprises because your input array is changed.
    weights: None or ndarray(T, )
        weights assigned to each trajectory point. If None, all data points have weight one.
        If ndarray, each data point is assigned a separate weight.
    sparse_mode : str
        one of:
            * 'dense' : always use dense mode
            * 'sparse' : always use sparse mode if possible
            * 'auto' : automatic
    sparse_tol: float
        Threshold for considering column to be zero in order to save computing
        effort when the data is sparse or almost sparse.
        If max(abs(X[:, i])) < sparse_tol, then row i (and also column i if Y
        is not given) of the covariance matrix will be set to zero. If Y is
        given and max(abs(Y[:, i])) < sparse_tol, then column i of the
        covariance matrix will be set to zero.
    column_selection: ndarray(k, dtype=int) or None
        Indices of those columns that are to be computed. If None, all columns are computed.
    diag_only: bool
        If True, the computation is restricted to the diagonal entries (autocorrelations) only.

    Returns
    -------
    w : float
        statistical weight
    s : ndarray (M)
        sum
    C : ndarray (M, M)
        unnormalized covariance matrix

    """
    # Check consistency of inputs:
    if weights is not None:
        assert X.shape[0] == weights.shape[0], 'X and weights_x must have equal length'
    # diag_only is only implemented for dense mode
    if diag_only and sparse_mode is not 'dense':
        if sparse_mode is 'sparse':
            import warnings
            warnings.warn('Computing diagonal entries only is not implemented for sparse mode. Switching to dense mode.')
        sparse_mode = 'dense'
    # sparsify
    X0, mask_X, xconst = _sparsify(X, remove_mean=remove_mean, modify_data=modify_data,
                                   sparse_mode=sparse_mode, sparse_tol=sparse_tol)
    is_sparse = mask_X is not None
    # copy / convert
    # TODO: do we need to copy xconst?
    X0, xconst = _copy_convert(X0, const=xconst, remove_mean=remove_mean,
                               copy=is_sparse or (remove_mean and not modify_data))
    # sum / center
    w, sx, sx0_centered = _sum(X0, xmask=mask_X, xconst=xconst, symmetric=False, remove_mean=remove_mean,
                               weights=weights)
    if remove_mean:
        _center(X0, w, sx, mask=mask_X, const=xconst, inplace=True)  # fast in-place centering
    # TODO: we could make a second const check here. If after summation not enough zeros have appeared in the
    # TODO: consts, we switch back to dense treatment here.
    # compute covariance matrix
    if column_selection is not None:
        if is_sparse:
            Xk = X[:, column_selection]
            mask_Xk = mask_X[column_selection]
            X0k = Xk[:, mask_Xk]
            xksum = sx0_centered[column_selection]
            xkconst = Xk[0, ~mask_Xk]
            X0k, xkconst = _copy_convert(X0k, const=xkconst, remove_mean=remove_mean,
                                         copy=True)
            C = _M2(X0, X0k, mask_X=mask_X, mask_Y=mask_Xk, xsum=sx0_centered, xconst=xconst, ysum=xksum, yconst=xkconst,
                    weights=weights)
        else:
            X0k = X0[:, column_selection]
            C = _M2(X0, X0k, mask_X=mask_X, mask_Y=mask_X, xsum=sx0_centered, xconst=xconst,
                    ysum=sx0_centered[column_selection], yconst=xconst, weights=weights)
    else:
        C = _M2(X0, X0, mask_X=mask_X, mask_Y=mask_X, xsum=sx0_centered, xconst=xconst, ysum=sx0_centered, yconst=xconst,
                weights=weights, diag_only=diag_only)
    return w, sx, C


def moments_XXXY(X, Y, remove_mean=False, symmetrize=False, weights=None,
                 modify_data=False, sparse_mode='auto', sparse_tol=0.0,
                 column_selection=None, diag_only=False):
    """ Computes the first two unnormalized moments of X and Y

    If symmetrize is False, computes

    .. math:
        s_x  &=& \sum_t x_t
        s_y  &=& \sum_t y_t
        C_XX &=& X^\top X
        C_XY &=& X^\top Y

    If symmetrize is True, computes

    .. math:
        s_x = s_y &=& \frac{1}{2} \sum_t(x_t + y_t)
        C_XX      &=& \frac{1}{2} (X^\top X + Y^\top Y)
        C_XY      &=& \frac{1}{2} (X^\top Y + Y^\top X)

    while exploiting zero or constant columns in the data matrix.

    Parameters
    ----------
    X : ndarray (T, M)
        Data matrix
    Y : ndarray (T, N)
        Second data matrix
    remove_mean : bool
        True: remove column mean from the data, False: don't remove mean.
    symmetrize : bool
        Computes symmetrized means and moments (see above)
    weights : None or ndarray(T, )
        weights assigned to each trajectory point of X. If None, all data points have weight one.
        If ndarray, each data point is assigned a separate weight.
    time_lagged : bool,
        indicates that Y is a time-lagged version of X.
    modify_data : bool
        If remove_mean=True, the mean will be removed in the data matrix X,
        without creating an independent copy. This option is faster but might
        lead to surprises because your input array is changed.
    sparse_mode : str
        one of:
            * 'dense' : always use dense mode
            * 'sparse' : always use sparse mode if possible
            * 'auto' : automatic
    sparse_tol: float
        Threshold for considering column to be zero in order to save computing
        effort when the data is sparse or almost sparse.
        If max(abs(X[:, i])) < sparse_tol, then row i (and also column i if Y
        is not given) of the covariance matrix will be set to zero. If Y is
        given and max(abs(Y[:, i])) < sparse_tol, then column i of the
        covariance matrix will be set to zero.
    column_selection: ndarray(k, dtype=int) or None
        Indices of those columns that are to be computed. If None, all columns are computed.
    diag_only: bool
        If True, the computation is restricted to the diagonal entries (autocorrelations) only.

    Returns
    -------
    w : float
        statistical weight
    s_x : ndarray (M)
        x-sum
    s_y : ndarray (N)
        y-sum
    C_XX : ndarray (M, M)
        unnormalized covariance matrix of X
    C_XY : ndarray (M, N)
        unnormalized covariance matrix of XY

    """
    # Check consistency of inputs:
    if Y is not None:
        assert Y.shape[0] == X.shape[0], 'X and Y must have equal length.'
    if weights is not None:
        assert X.shape[0] == weights.shape[0], 'X and weights_x must have equal length'
    # diag_only is only implemented for dense mode
    if diag_only and sparse_mode is not 'dense':
        if sparse_mode is 'sparse':
            import warnings
            warnings.warn('Computing diagonal entries only is not implemented for sparse mode. Switching to dense mode.')
        sparse_mode = 'dense'
    if diag_only and X.shape[1] != Y.shape[1]:
        raise ValueError('Computing diagonal entries only does not make sense for rectangular covariance matrix.')
    # sparsify
    X0, mask_X, xconst, Y0, mask_Y, yconst = _sparsify_pair(X, Y, remove_mean=remove_mean, modify_data=modify_data,
                                                            symmetrize=symmetrize, sparse_mode=sparse_mode, sparse_tol=sparse_tol)
    is_sparse = mask_X is not None and mask_Y is not None
    # copy / convert
    copy = is_sparse or (remove_mean and not modify_data)
    X0, xconst = _copy_convert(X0, const=xconst, remove_mean=remove_mean, copy=copy)
    Y0, yconst = _copy_convert(Y0, const=yconst, remove_mean=remove_mean, copy=copy)
    # sum / center
    w, sx, sx_centered, sy, sy_centered = _sum(X0, xmask=mask_X, xconst=xconst, Y=Y0, ymask=mask_Y, yconst=yconst,
                                               symmetric=symmetrize, remove_mean=remove_mean, weights=weights)
    if remove_mean:
        _center(X0, w, sx, mask=mask_X, const=xconst, inplace=True)  # fast in-place centering
        _center(Y0, w, sy, mask=mask_Y, const=yconst, inplace=True)  # fast in-place centering

    if symmetrize:
        Cxx, Cxy = _M2_symmetric(X0, Y0, mask_X=mask_X, mask_Y=mask_Y,
                                 xsum=sx_centered, xconst=xconst, ysum=sy_centered, yconst=yconst, weights=weights,
                                 column_selection=column_selection, diag_only=diag_only)
    else:
        if column_selection is not None:
            if is_sparse:
                Xk = X[:, column_selection]
                mask_Xk = mask_X[column_selection]
                X0k = Xk[:, mask_Xk]
                xksum = sx_centered[column_selection]
                xkconst = Xk[0, ~mask_Xk]
                X0k, xkconst = _copy_convert(X0k, const=xkconst, remove_mean=remove_mean,
                                             copy=True)

                Yk = Y[:, column_selection]
                mask_Yk = mask_Y[column_selection]
                Y0k = Yk[:, mask_Yk]
                yksum = sy_centered[column_selection]
                ykconst = Yk[0, ~mask_Yk]
                Y0k, ykconst = _copy_convert(Y0k, const=ykconst, remove_mean=remove_mean,
                                             copy=True)

                Cxx = _M2(X0, X0k, mask_X=mask_X, mask_Y=mask_Xk, xsum=sx_centered, xconst=xconst, ysum=xksum, yconst=xkconst,
                        weights=weights)
                Cxy = _M2(X0, Y0k, mask_X=mask_X, mask_Y=mask_Yk, xsum=sx_centered, xconst=xconst, ysum=yksum, yconst=ykconst,
                        weights=weights)
            else:
                X0k = X0[:, column_selection]
                Y0k = Y0[:, column_selection]
                Cxx = _M2(X0, X0k, mask_X=mask_X, mask_Y=mask_X, xsum=sx_centered, xconst=xconst,
                        ysum=sx_centered[column_selection], yconst=xconst, weights=weights)
                Cxy = _M2(X0, Y0k, mask_X=mask_X, mask_Y=mask_Y, xsum=sx_centered, xconst=xconst,
                        ysum=sy_centered[column_selection], yconst=yconst, weights=weights)
        else:
            Cxx = _M2(X0, X0, mask_X=mask_X, mask_Y=mask_X, xsum=sx_centered, xconst=xconst, ysum=sx_centered, yconst=xconst,
                      weights=weights, diag_only=diag_only)
            Cxy = _M2(X0, Y0, mask_X=mask_X, mask_Y=mask_Y, xsum=sx_centered, xconst=xconst, ysum=sy_centered, yconst=yconst,
                      weights=weights, diag_only=diag_only)

    return w, sx, sy, Cxx, Cxy


def moments_block(X, Y, remove_mean=False, modify_data=False,
                  sparse_mode='auto', sparse_tol=0.0,
                  column_selection=None, diag_only=False):
    """ Computes the first two unnormalized moments of X and Y

    Computes

    .. math:
        s_x  &=& \sum_t x_t
        s_y  &=& \sum_t y_t
        C_XX &=& X^\top X
        C_XY &=& X^\top Y
        C_YX &=& Y^\top X
        C_YY &=& Y^\top Y

    while exploiting zero or constant columns in the data matrix.

    Parameters
    ----------
    X : ndarray (T, M)
        Data matrix
    Y : ndarray (T, N)
        Second data matrix
    remove_mean : bool
        True: remove column mean from the data, False: don't remove mean.
    modify_data : bool
        If remove_mean=True, the mean will be removed in the data matrix X,
        without creating an independent copy. This option is faster but might
        lead to surprises because your input array is changed.
    sparse_mode : str
        one of:
            * 'dense' : always use dense mode
            * 'sparse' : always use sparse mode if possible
            * 'auto' : automatic
    sparse_tol: float
        Threshold for considering column to be zero in order to save computing
        effort when the data is sparse or almost sparse.
        If max(abs(X[:, i])) < sparse_tol, then row i (and also column i if Y
        is not given) of the covariance matrix will be set to zero. If Y is
        given and max(abs(Y[:, i])) < sparse_tol, then column i of the
        covariance matrix will be set to zero.
    column_selection: ndarray(k, dtype=int) or None
        Indices of those columns that are to be computed. If None, all columns are computed.
    diag_only: bool
        If True, the computation is restricted to the diagonal entries (autocorrelations) only.

    Returns
    -------
    w : float
        statistical weight of this estimation
    s : [ndarray (M), ndarray (M)]
        list of two elements with s[0]=sx and s[1]=sy
    C : [[ndarray(M,M), ndarray(M,N)], [ndarray(N,M),ndarray(N,N)]]
        list of two lists with two elements.
        C[0,0] = Cxx, C[0,1] = Cxy, C[1,0] = Cyx, C[1,1] = Cyy

    """
    # diag_only is only implemented for dense mode
    if diag_only and sparse_mode is not 'dense':
        if sparse_mode is 'sparse':
            import warnings
            warnings.warn('Computing diagonal entries only is not implemented for sparse mode. Switching to dense mode.')
        sparse_mode = 'dense'
    # sparsify
    X0, mask_X, xconst = _sparsify(X, sparse_mode=sparse_mode, sparse_tol=sparse_tol)
    Y0, mask_Y, yconst = _sparsify(Y, sparse_mode=sparse_mode, sparse_tol=sparse_tol)
    is_sparse = mask_X is not None and mask_Y is not None
    # copy / convert
    copy = is_sparse or (remove_mean and not modify_data)
    X0, xconst = _copy_convert(X0, const=xconst, copy=copy)
    Y0, yconst = _copy_convert(Y0, const=yconst, copy=copy)
    # sum / center
    w, sx, sx_centered, sy, sy_centered = _sum(X0, xmask=mask_X, xconst=xconst, Y=Y0, ymask=mask_Y, yconst=yconst,
                                               symmetric=False, remove_mean=remove_mean)
    if remove_mean:
        _center(X0, w, sx, mask=mask_X, const=xconst, inplace=True)  # fast in-place centering
        _center(Y0, w, sy, mask=mask_Y, const=yconst, inplace=True)  # fast in-place centering

    if column_selection is not None:
        if is_sparse:
            Xk = X[:, column_selection]
            mask_Xk = mask_X[column_selection] if mask_X is not None else mask_X
            X0k = Xk[:, mask_Xk]
            xksum = sx_centered[column_selection]
            xkconst = Xk[0, ~mask_Xk]
            X0k, xkconst = _copy_convert(X0k, const=xkconst, remove_mean=remove_mean,
                                         copy=True)

            Yk = Y[:, column_selection]
            mask_Yk = mask_Y[column_selection] if mask_Y is not None else mask_Y
            Y0k = Yk[:, mask_Yk]
            yksum = sy_centered[column_selection]
            ykconst = Yk[0, ~mask_Yk]
            Y0k, ykconst = _copy_convert(Y0k, const=ykconst, remove_mean=remove_mean,
                                         copy=True)

            Cxx = _M2(X0, X0k, mask_X=mask_X, mask_Y=mask_Xk,
                      xsum=sx_centered, xconst=xconst, ysum=xksum, yconst=xkconst)
            Cxy = _M2(X0, Y0k, mask_X=mask_X, mask_Y=mask_Yk,
                      xsum=sx_centered, xconst=xconst, ysum=yksum, yconst=ykconst)
            Cyx = _M2(Y0, X0k, mask_X=mask_Y, mask_Y=mask_Xk,
                      xsum=sy_centered, xconst=yconst, ysum=xksum, yconst=xkconst)
            Cyy = _M2(Y0, Y0k, mask_X=mask_Y, mask_Y=mask_Yk,
                      xsum=sy_centered, xconst=yconst, ysum=yksum, yconst=ykconst)
        else:
            X0k = X0[:, column_selection]
            Y0k = Y0[:, column_selection]
            Cxx = _M2(X0, X0k, mask_X=mask_X, mask_Y=mask_X,
                      xsum=sx_centered, xconst=xconst,
                      ysum=sx_centered[column_selection], yconst=xconst)
            Cxy = _M2(X0, Y0k, mask_X=mask_X, mask_Y=mask_Y,
                      xsum=sx_centered, xconst=xconst,
                      ysum=sy_centered[column_selection], yconst=yconst)
            Cyx = _M2(Y0, X0k, mask_X=mask_Y, mask_Y=mask_X,
                      xsum=sy_centered, xconst=yconst,
                      ysum=sx_centered[column_selection], yconst=xconst)
            Cyy = _M2(Y0, Y0k, mask_X=mask_Y, mask_Y=mask_Y,
                      xsum=sy_centered, xconst=yconst,
                      ysum=sy_centered[column_selection], yconst=yconst)
    else:
        Cxx = _M2(X0, X0, mask_X=mask_X, mask_Y=mask_X,
                  xsum=sx_centered, xconst=xconst, ysum=sx_centered, yconst=xconst,
                  diag_only=diag_only)
        Cxy = _M2(X0, Y0, mask_X=mask_X, mask_Y=mask_Y,
                  xsum=sx_centered, xconst=xconst, ysum=sy_centered, yconst=yconst,
                  diag_only=diag_only)
        Cyx = Cxy.T
        Cyy = _M2(Y0, Y0, mask_X=mask_Y, mask_Y=mask_Y,
                  xsum=sy_centered, xconst=yconst, ysum=sy_centered, yconst=yconst,
                  diag_only=diag_only)

    return w, (sx, sy), ((Cxx, Cxy), (Cyx, Cyy))


def covar(X, remove_mean=False, modify_data=False, weights=None, sparse_mode='auto', sparse_tol=0.0):
    """ Computes the covariance matrix of X

    Computes

    .. math:
        C_XX &=& X^\top X

    while exploiting zero or constant columns in the data matrix.
    WARNING: Directly use moments_XX if you can. This function does an additional
    constant-matrix multiplication and does not return the mean.

    Parameters
    ----------
    X : ndarray (T, M)
        Data matrix
    remove_mean : bool
        True: remove column mean from the data, False: don't remove mean.
    modify_data : bool
        If remove_mean=True, the mean will be removed in the data matrix X,
        without creating an independent copy. This option is faster but might
        lead to surprises because your input array is changed.
    weights : None or ndarray(T, )
        weights assigned to each trajectory point of X. If None, all data points have weight one.
        If ndarray, each data point is assigned a separate weight.
    sparse_mode : str
        one of:
            * 'dense' : always use dense mode
            * 'sparse' : always use sparse mode if possible
            * 'auto' : automatic
    sparse_tol: float
        Threshold for considering column to be zero in order to save computing
        effort when the data is sparse or almost sparse.
        If max(abs(X[:, i])) < sparse_tol, then row i (and also column i if Y
        is not given) of the covariance matrix will be set to zero. If Y is
        given and max(abs(Y[:, i])) < sparse_tol, then column i of the
        covariance matrix will be set to zero.

    Returns
    -------
    C_XX : ndarray (M, M)
        Covariance matrix of X

    See also
    --------
    moments_XX

    """
    w, s, M = moments_XX(X, remove_mean=remove_mean, weights=weights, modify_data=modify_data,
                         sparse_mode=sparse_mode, sparse_tol=sparse_tol)
    return M / float(w)


def covars(X, Y, remove_mean=False, modify_data=False, symmetrize=False, weights=None, sparse_mode='auto',
           sparse_tol=0.0):
    """ Computes the covariance and cross-covariance matrix of X and Y

    If symmetrize is False, computes

    .. math:
        C_XX &=& X^\top X
        C_XY &=& X^\top Y

    If symmetrize is True, computes

    .. math:
        C_XX      &=& \frac{1}{2} (X^\top X + Y^\top Y)
        C_XY      &=& \frac{1}{2} (X^\top Y + Y^\top X)

    while exploiting zero or constant columns in the data matrix.
    WARNING: Directly use moments_XXXY if you can. This function does an additional
    constant-matrix multiplication and does not return the mean.

    Parameters
    ----------
    X : ndarray (T, M)
        Data matrix
    Y : ndarray (T, N)
        Second data matrix
    remove_mean : bool
        True: remove column mean from the data, False: don't remove mean.
    modify_data : bool
        If remove_mean=True, the mean will be removed in the data matrix X,
        without creating an independent copy. This option is faster but might
        lead to surprises because your input array is changed.
    symmetrize : bool
        Computes symmetrized means and moments (see above)
    weights : None or ndarray(T, )
        weights assigned to each trajectory point of X. If None, all data points have weight one.
        If ndarray, each data point is assigned a separate weight.
    sparse_mode : str
        one of:
            * 'dense' : always use dense mode
            * 'sparse' : always use sparse mode if possible
            * 'auto' : automatic
    sparse_tol: float
        Threshold for considering column to be zero in order to save computing
        effort when the data is sparse or almost sparse.
        If max(abs(X[:, i])) < sparse_tol, then row i (and also column i if Y
        is not given) of the covariance matrix will be set to zero. If Y is
        given and max(abs(Y[:, i])) < sparse_tol, then column i of the
        covariance matrix will be set to zero.

    Returns
    -------
    C_XX : ndarray (M, M)
        Covariance matrix of X
    C_XY : ndarray (M, N)
        Covariance matrix of XY

    See also
    --------
    moments_XXXY

    """
    w, sx, sy, Mxx, Mxy = moments_XXXY(X, Y, remove_mean=remove_mean, modify_data=modify_data, weights=weights,
                                       symmetrize=symmetrize, sparse_mode=sparse_mode, sparse_tol=sparse_tol)
    return Mxx / float(w), Mxy / float(w)




# This file is part of PyEMMA.
#
# Copyright (c) 2014-2016 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
Created on 22.05.2015

@author: marscher
'''

import unittest
import numpy as np

from pyemma.plots.networks import plot_flux, plot_markov_model, plot_network
from pyemma.msm import tpt, MSM
import matplotlib
import matplotlib.pyplot as plt


class TestNetworkPlot(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # 5-state toy system
        cls.P = np.array([[0.8, 0.15, 0.05, 0.0, 0.0],
                          [0.1, 0.75, 0.05, 0.05, 0.05],
                          [0.05, 0.1, 0.8, 0.0, 0.05],
                          [0.0, 0.2, 0.0, 0.8, 0.0],
                          [0.0, 0.02, 0.02, 0.0, 0.96]])
        cls.A = [0]
        cls.B = [4]
        cls.msm = MSM(cls.P)

        cls.P_mfpt = np.zeros_like(cls.P)
        for ii in np.arange(cls.P.shape[0]):
            for jj in np.arange(cls.P.shape[1]):
                cls.P_mfpt[ii,jj] = cls.msm.mfpt([ii], [jj])
        return cls

    def test_flux(self):
        r = tpt(self.msm, self.A, self.B)
        fig, pos = plot_flux(r)
        assert type(fig) is matplotlib.figure.Figure
        # matplotlib.pyplot.show(fig)
        # x values should be close to committor
        np.testing.assert_allclose(pos[:,0], r.committor)

    def test_plot_markov_model(self):
        fig, pos = plot_markov_model(self.P)
        assert type(fig) is matplotlib.figure.Figure

    def test_forced_no_arrows_labels(self):
        fig, pos = plot_markov_model(self.P, arrow_labels=None)
        assert type(fig) is matplotlib.figure.Figure

    def test_numeric_arrow_labels(self):
        fig, pos = plot_markov_model(self.P, arrow_labels=self.P)
        assert type(fig) is matplotlib.figure.Figure

    def test_alphanumeric_arrow_labels(self):
        fig, pos = plot_markov_model(
            self.P, arrow_labels=self.P_mfpt, arrow_label_format='mfpts = %f frames')
        assert type(fig) is matplotlib.figure.Figure

    def test_alphanumeric_arrow_labels_using_ax(self):
        orig_fig, ax = plt.subplots()
        fig, pos = plot_markov_model(
            self.P, arrow_labels=self.P_mfpt, arrow_label_format='mfpts = %f frames', ax=ax)
        assert type(fig) is matplotlib.figure.Figure
        assert fig == orig_fig

    def test_string_arrow_labels_using_ax(self):
        orig_fig, ax = plt.subplots()
        labels = np.array([['A2A', 'A2B'],
                           ['B2A', 'B2B']]
                  )
        fig, pos = plot_network(
            self.P[:2,:2], arrow_labels=labels, ax=ax)
        assert type(fig) is matplotlib.figure.Figure

    def test_state_labels_network(self):
        state_labels = [str(i) for i in np.arange(len(self.P))]
        plot_network(self.P, state_labels=state_labels)

        with self.assertRaises(ValueError):
            plot_network(self.P, state_labels=state_labels[:2])

        state_labels = 'auto'
        plot_network(self.P, state_labels=state_labels)

    def test_state_labels_flux(self):
        """ ensure our labels show up in the plot"""
        flux = tpt(self.msm, [0,1], [2,4])
        labels = ['foo', '0', '1', '2', 'bar']
        fig, pos = plot_flux(flux, state_labels=labels)
        labels_in_fig = np.array([text.get_text() for text in fig.axes[0].texts])

        for l in labels:
            self.assertEqual((labels_in_fig == l).sum(), 1)

    def test_state_labels_flux_auto(self):
        """ ensure auto generated labels show up in the plot"""
        A = [0,1]
        B = [2,4]
        flux = tpt(self.msm, A, B)
        fig, pos = plot_flux(flux, state_labels='auto')
        labels_in_fig = np.array([text.get_text() for text in fig.axes[0].texts])
        self.assertEqual((labels_in_fig == "A").sum(), len(A))
        self.assertEqual((labels_in_fig == "B").sum(), len(B))

if __name__ == "__main__":
    unittest.main()



# This file is part of PyEMMA.
#
# Copyright (c) 2015, 2014 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
import unittest

import pkg_resources
import mdtraj
import numpy as np
import pyemma.coordinates as coor
from pyemma.coordinates.data.fragmented_trajectory_reader import FragmentedTrajectoryReader


class TestFragmentedTrajectory(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        d = np.array([[i] for i in range(0, 100)]) # np.atleast_2d(np.arange(100))
        cls.d = d
        return cls

    def test_full_trajectory(self):
        reader = FragmentedTrajectoryReader([self.d, self.d])
        reader.chunksize = 0
        expected = np.vstack((self.d, self.d))
        output = reader.get_output(stride=1)[0]
        np.testing.assert_array_almost_equal(expected, output)

    def test_full_trajectory_random_access(self):
        reader = FragmentedTrajectoryReader([self.d, self.d])
        indices = np.asarray([[0, 1], [0, 3], [0, 3], [0, 99], [0, 100], [0, 199]])
        out = reader.get_output(stride=indices, chunk=0)
        np.testing.assert_array_equal(np.array(out).squeeze(), np.array([1, 3, 3, 99, 0, 99]))

    def test_chunked_trajectory_random_access(self):
        reader = FragmentedTrajectoryReader([self.d, self.d])
        indices = np.asarray([[0, 1], [0, 3], [0, 3], [0, 99], [0, 100], [0, 199]])
        out = reader.get_output(stride=indices, chunk=1)
        np.testing.assert_array_equal(np.array(out).squeeze(), np.array([1,3,3,99,0,99]))

    def test_full_trajectory_stridden(self):
        for stride in [1, 3, 5, 7, 13, 20]:
            reader = FragmentedTrajectoryReader([self.d, self.d])
            reader.chunksize = 0
            expected = np.vstack((self.d, self.d))[::stride]
            out = reader.get_output(stride=stride)[0]
            np.testing.assert_array_almost_equal(expected, out, err_msg="Failed for stride=%s" % stride)

    def test_full_trajectory_stridden_with_lag(self):
        reader = FragmentedTrajectoryReader([self.d, self.d])
        data = np.vstack((self.d, self.d))
        for lag in [1, 5, 7]:
            for stride in [1, 3, 5, 7, 13, 20]:
                reader.chunksize = 0

                X, Y = None, None
                # not chunked
                for itraj, X, Y in reader.iterator(stride=stride, lag=lag):
                    pass

                np.testing.assert_array_almost_equal(data[::stride][0:len(Y)], X)
                np.testing.assert_array_almost_equal(data[lag::stride], Y, err_msg='lag={lag}, stride={stride}'.format(stride=stride, lag=lag))

    def test_fragmented_xtc(self):
        from pyemma.coordinates.tests.util import create_traj

        top_file = pkg_resources.resource_filename(__name__, 'data/test.pdb')
        trajfiles = []
        for _ in range(3):
            f, _, _ = create_traj(top_file)
            trajfiles.append(f)
        try:
            # three trajectories: one consisting of all three, one consisting of the first,
            # one consisting of the first and the last
            source = coor.source([trajfiles, [trajfiles[0]], [trajfiles[0], trajfiles[2]]], top=top_file)
            source.chunksize = 1000

            out = source.get_output(stride=1)
            trajs = [mdtraj.load(trajfiles[i], top=top_file).xyz.reshape(-1,9) for i in range(0,3)]

            np.testing.assert_equal(out[0], np.vstack(trajs))
            np.testing.assert_equal(out[1], trajs[0])
            np.testing.assert_equal(out[2], np.vstack((trajs[0], trajs[2])))
        finally:
            for t in trajfiles:
                try:
                    os.unlink(t)
                except EnvironmentError:
                    pass

    def test_multiple_input_trajectories_random_access(self):
        indices = np.asarray([
            [0, 1], [0, 3], [0, 3], [0, 99], [0, 100], [0, 199],
            [1, 0], [1, 5], [1, 99],
            [2, 5], [2, 7], [2, 23]
        ])
        expected = [np.array([1, 3, 3, 99, 0, 99]), np.array([0, 5, 99]), np.array([5, 7, 23])]
        for chunk_size in [0, 1, 3, 5, 13]:
            reader = FragmentedTrajectoryReader([[self.d, self.d], self.d, [self.d, self.d]])
            out_full_trajectory_mode = reader.get_output(chunk=chunk_size, stride=indices)
            for i in range(3):
                np.testing.assert_array_equal(expected[i], out_full_trajectory_mode[i].squeeze())

    def test_multiple_input_trajectories(self):
        reader = FragmentedTrajectoryReader([[self.d, self.d], self.d, [self.d, self.d]])
        reader.chunksize = 37
        out = reader.get_output()
        reader.chunksize = 0
        out2 = reader.get_output()
        expected0_2 = np.vstack((self.d, self.d))
        for itraj in range(0, 3):
            np.testing.assert_array_almost_equal(out[itraj], out2[itraj])
        np.testing.assert_array_almost_equal(out[0], expected0_2)
        np.testing.assert_array_almost_equal(out[1], self.d)
        np.testing.assert_array_almost_equal(out[2], expected0_2)

    def test_chunked_trajectory_with_lag(self):
        data = np.vstack((self.d, self.d))
        reader = FragmentedTrajectoryReader([self.d, self.d])
        for lag in [0, 1, 3]:
            for stride in [1, 3, 5]:
                for chunksize in [1, 34, 53, 72]:
                    reader.chunksize = chunksize
                    if lag > 0:
                        collected = None
                        collected_lagged = None
                        for itraj, X, Y in reader.iterator(stride=stride, lag=lag):
                            collected = X if collected is None else np.vstack((collected, X))
                            collected_lagged = Y if collected_lagged is None else np.vstack((collected_lagged, Y))
                        np.testing.assert_array_almost_equal(data[::stride][0:len(collected_lagged)], collected,
                                                             err_msg="lag={}, stride={}, cs={}".format(
                                                                 lag, stride, chunksize
                                                             ))
                        np.testing.assert_array_almost_equal(data[lag::stride], collected_lagged)
                    else:
                        collected = None
                        for itraj, X in reader.iterator(stride=stride):
                            collected = X if collected is None else np.vstack((collected, X))
                        np.testing.assert_array_almost_equal(data[::stride], collected)

    def test_index_to_reader_index(self):
        reader = FragmentedTrajectoryReader([self.d, self.d])
        assert (0, 0) == reader._index_to_reader_index(0, 0), "first frame is first frame of first reader"
        assert (0, 1) == reader._index_to_reader_index(1, 0), "second frame is second frame of first reader"
        assert (1, 0) == reader._index_to_reader_index(100, 0), "101'st frame is first frame of second reader"
        assert (1, 1) == reader._index_to_reader_index(101, 0), "102'nd frame is second frame of second reader"
        with self.assertRaises(ValueError):
            reader._index_to_reader_index(-1, 0)
        with self.assertRaises(ValueError):
            reader._index_to_reader_index(200, 0)

    def test_cols(self):
        dim = 5
        arr = np.arange(60).reshape(-1, dim)
        data = [(arr, arr), arr, (arr, arr, arr)]
        reader = FragmentedTrajectoryReader(data)
        cols = (0, 3)
        for itraj, x in reader.iterator(chunk=0, return_trajindex=True, cols=cols):
            if isinstance(data[itraj], tuple):
                syn_traj = np.concatenate(data[itraj])
            else:
                syn_traj = data[itraj]
            np.testing.assert_equal(x, syn_traj[:, cols])

    def test_raise_different_dims(self):
        data = [self.d, np.array([[1,2,3], [4,5,6]])]
        with self.assertRaises(ValueError):
            FragmentedTrajectoryReader(data)

    def test_with_save_traj(self):
        path = pkg_resources.resource_filename(__name__, 'data') + os.path.sep

        pdb_file = os.path.join(path, 'bpti_ca.pdb')
        traj_files = [
            os.path.join(path, 'bpti_001-033.xtc'),
            os.path.join(path, 'bpti_034-066.xtc'),
            os.path.join(path, 'bpti_067-100.xtc')
        ]

        source_frag = coor.source([traj_files], top=pdb_file)
        full_data = source_frag.get_output()[0]
        last_frame_fragment_0 = [0,32]
        first_frame_fragment_1 = [0,33]
        first_frame_fragment_2 = [0,66]

        reshape = lambda f: f.xyz.reshape((f.xyz.shape[0],f.xyz.shape[1] * f.xyz.shape[2])).squeeze()

        # Frames in the first fragment:
        frames = coor.save_traj(source_frag, [last_frame_fragment_0], None)
        np.testing.assert_equal(reshape(frames), full_data[32])

        # Frames the first and second fragments
        frames = coor.save_traj(source_frag, [last_frame_fragment_0, first_frame_fragment_1], None)
        np.testing.assert_equal(reshape(frames), full_data[np.array([32, 33])])

        # Frames only in the second fragment
        frames = coor.save_traj(source_frag, [first_frame_fragment_1], None)
        np.testing.assert_equal(reshape(frames), full_data[33])

        # Frames only in the second and third fragment
        frames = coor.save_traj(source_frag, [first_frame_fragment_1, first_frame_fragment_2], None)
        np.testing.assert_equal(reshape(frames), full_data[np.array([33, 66])])




# This file is part of PyEMMA.
#
# Copyright (c) 2015, 2014 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
Created on 19.01.2015

@author: marscher
'''

import os
import tempfile
import unittest
import mdtraj
import numpy as np

from mdtraj.core.trajectory import Trajectory
from mdtraj.core.element import hydrogen, oxygen
from mdtraj.core.topology import Topology

from pyemma.coordinates.clustering.uniform_time import UniformTimeClustering
from pyemma.coordinates.pipelines import Discretizer
from pyemma.coordinates.data.data_in_memory import DataInMemory
from pyemma.coordinates.api import cluster_kmeans, pca, source



def create_water_topology_on_disc(n):
    topfile = tempfile.mktemp('.pdb')
    top = Topology()
    chain = top.add_chain()

    for i in range(n):
        res = top.add_residue('r%i' % i, chain)
        h1 = top.add_atom('H', hydrogen, res)
        o = top.add_atom('O', oxygen, res)
        h2 = top.add_atom('H', hydrogen, res)
        top.add_bond(h1, o)
        top.add_bond(h2, o)

    xyz = np.zeros((n * 3, 3))
    Trajectory(xyz, top).save_pdb(topfile)
    return topfile


def create_traj_on_disc(topfile, n_frames, n_atoms):
    fn = tempfile.mktemp('.xtc')
    xyz = np.random.random((n_frames, n_atoms, 3))
    t = mdtraj.load(topfile)
    t.xyz = xyz
    t.time = np.arange(n_frames)
    t.save(fn)
    return fn


class TestDiscretizer(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        c = super(TestDiscretizer, cls).setUpClass()
        # create a fake trajectory which has 2 atoms and coordinates are just a range
        # over all frames.
        cls.n_frames = 1000
        cls.n_residues = 30
        cls.topfile = create_water_topology_on_disc(cls.n_residues)

        # create some trajectories
        t1 = create_traj_on_disc(
            cls.topfile, cls.n_frames, cls.n_residues * 3)

        t2 = create_traj_on_disc(
            cls.topfile, cls.n_frames, cls.n_residues * 3)

        cls.trajfiles = [t1, t2]

        cls.dest_dir = tempfile.mkdtemp()

        return c

    @classmethod
    def tearDownClass(cls):
        """delete temporary files"""
        os.unlink(cls.topfile)
        for f in cls.trajfiles:
            os.unlink(f)

        import shutil
        shutil.rmtree(cls.dest_dir, ignore_errors=True)

    def test(self):
        reader = source(self.trajfiles, top=self.topfile)
        pcat = pca(dim=2)

        n_clusters = 2
        clustering = UniformTimeClustering(n_clusters=n_clusters)

        D = Discretizer(reader, transform=pcat, cluster=clustering)
        D.parametrize()

        self.assertEqual(len(D.dtrajs), len(self.trajfiles))

        for dtraj in clustering.dtrajs:
            unique = np.unique(dtraj)
            self.assertEqual(unique.shape[0], n_clusters)

    def test_with_data_in_mem(self):
        import pyemma.coordinates as api

        data = [np.random.random((100, 50)),
                np.random.random((103, 50)),
                np.random.random((33, 50))]
        reader = source(data)
        assert isinstance(reader, DataInMemory)

        tpca = api.pca(dim=2)

        n_centers = 10
        km = api.cluster_kmeans(k=n_centers)

        disc = api.discretizer(reader, tpca, km)
        disc.parametrize()

        dtrajs = disc.dtrajs
        for dtraj in dtrajs:
            n_states = np.max((np.unique(dtraj)))
            self.assertGreaterEqual(n_centers - 1, n_states,
                                    "dtraj has more states than cluster centers")

    def test_save_dtrajs(self):
        reader = source(self.trajfiles, top=self.topfile)
        cluster = cluster_kmeans(k=2)
        d = Discretizer(reader, cluster=cluster)
        d.parametrize()
        d.save_dtrajs(output_dir=self.dest_dir)
        dtrajs = os.listdir(self.dest_dir)


if __name__ == "__main__":
    unittest.main()



"""
Interfacing functions and classing to Serialization.

Derive from SerializableMixIn to make your class serializable. If you need to patch an old version of your class,
you need the Modifications class.

"""

from .serialization import SerializableMixIn, Modifications

__all__ = ['SerializableMixIn', 'Modifications', 'load', 'list_models']


def load(filename, model_name='default'):
    from .serialization import SerializableMixIn
    return SerializableMixIn.load(file_name=filename, model_name=model_name)


load.__doc__ = SerializableMixIn.load.__doc__


def list_models(filename):
    """ Lists all models in given filename.

    Parameters
    ----------
    filename: str
        path to filename, where the model has been stored.

    Returns
    -------
    obj: dict
        A mapping by name and a comprehensive description like this:
        {model_name: {'repr' : 'string representation, 'created': 'human readable date', ...}
    """
    from .h5file import H5File
    with H5File(filename, mode='r') as f:
        return f.models_descriptive



# This file is part of PyEMMA.
#
# Copyright (c) 2015, 2014 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''
Created on 19.01.2015

@author: marscher
'''


import numpy as np
from pyemma._base.serialization.serialization import SerializableMixIn

from pyemma._ext.variational.solvers.direct import eig_corr
from pyemma._ext.variational.util import ZeroRankError
from pyemma.coordinates.estimation.covariance import LaggedCovariance
from pyemma.coordinates.transform._tica_base import TICABase, TICAModelBase
from pyemma.util.annotators import fix_docs
import warnings

__all__ = ['TICA']


@fix_docs
class TICA(TICABase, SerializableMixIn):
    r""" Time-lagged independent component analysis (TICA)"""
    __serialize_version = 0

    def __init__(self, lag, dim=-1, var_cutoff=0.95, kinetic_map=True, commute_map=False, epsilon=1e-6,
                 stride=1, skip=0, reversible=True, weights=None, ncov_max=float('inf')):
        r""" Time-lagged independent component analysis (TICA) [1]_, [2]_, [3]_.

        Parameters
        ----------
        lag : int
            lag time
        dim : int, optional, default -1
            Maximum number of significant independent components to use to reduce dimension of input data. -1 means
            all numerically available dimensions (see epsilon) will be used unless reduced by var_cutoff.
            Setting dim to a positive value is exclusive with var_cutoff.
        var_cutoff : float in the range [0,1], optional, default 0.95
            Determines the number of output dimensions by including dimensions until their cumulative kinetic variance
            exceeds the fraction subspace_variance. var_cutoff=1.0 means all numerically available dimensions
            (see epsilon) will be used, unless set by dim. Setting var_cutoff smaller than 1.0 is exclusive with dim
        kinetic_map : bool, optional, default True
            Eigenvectors will be scaled by eigenvalues. As a result, Euclidean distances in the transformed data
            approximate kinetic distances [4]_. This is a good choice when the data is further processed by clustering.
        commute_map : bool, optional, default False
            Eigenvector_i will be scaled by sqrt(timescale_i / 2). As a result, Euclidean distances in the transformed
            data will approximate commute distances [5]_.
        epsilon : float
            eigenvalue norm cutoff. Eigenvalues of C0 with norms <= epsilon will be
            cut off. The remaining number of eigenvalues define the size
            of the output.
        stride: int, optional, default = 1
            Use only every stride-th time step. By default, every time step is used.
        skip : int, default=0
            skip the first initial n frames per trajectory.
        reversible: bool, default=True
            symmetrize correlation matrices C_0, C_{\tau}.
        weights: object or list of ndarrays, optional, default = None
            * An object that allows to compute re-weighting factors to estimate equilibrium means and correlations from
              off-equilibrium data. The only requirement is that weights possesses a method weights(X), that accepts a
              trajectory X (np.ndarray(T, n)) and returns a vector of re-weighting factors (np.ndarray(T,)).
            * A list of ndarrays (ndim=1) specifies the weights for each frame of each trajectory.

        Notes
        -----
        Given a sequence of multivariate data :math:`X_t`, computes the mean-free
        covariance and time-lagged covariance matrix:

        .. math::

            C_0 &=      (X_t - \mu)^T (X_t - \mu) \\
            C_{\tau} &= (X_t - \mu)^T (X_{t + \tau} - \mu)

        and solves the eigenvalue problem

        .. math:: C_{\tau} r_i = C_0 \lambda_i(tau) r_i,

        where :math:`r_i` are the independent components and :math:`\lambda_i(tau)` are
        their respective normalized time-autocorrelations. The eigenvalues are
        related to the relaxation timescale by

        .. math:: t_i(tau) = -\tau / \ln |\lambda_i|.

        When used as a dimension reduction method, the input data is projected
        onto the dominant independent components.

        References
        ----------
        .. [1] Perez-Hernandez G, F Paul, T Giorgino, G De Fabritiis and F Noe. 2013.
           Identification of slow molecular order parameters for Markov model construction
           J. Chem. Phys. 139, 015102. doi:10.1063/1.4811489
        .. [2] Schwantes C, V S Pande. 2013.
           Improvements in Markov State Model Construction Reveal Many Non-Native Interactions in the Folding of NTL9
           J. Chem. Theory. Comput. 9, 2000-2009. doi:10.1021/ct300878a
        .. [3] L. Molgedey and H. G. Schuster. 1994.
           Separation of a mixture of independent signals using time delayed correlations
           Phys. Rev. Lett. 72, 3634.
        .. [4] Noe, F. and Clementi, C. 2015. Kinetic distance and kinetic maps from molecular dynamics simulation.
            J. Chem. Theory. Comput. doi:10.1021/acs.jctc.5b00553
        .. [5] Noe, F., Banisch, R., Clementi, C. 2016. Commute maps: separating slowly-mixing molecular configurations
           for kinetic modeling. J. Chem. Theory. Comput. doi:10.1021/acs.jctc.6b00762

        """
        super(TICA, self).__init__()
        if kinetic_map and commute_map:
            raise ValueError('Trying to use both kinetic_map and commute_map. Use either or.')
        if (kinetic_map or commute_map) and not reversible:
            kinetic_map = False
            commute_map = False
            warnings.warn("Cannot use kinetic_map or commute_map for non-reversible processes, both will be set to"
                          "False.")

        # this instance will be set by partial fit.
        self._covar = None

        self.dim = dim
        self.var_cutoff = var_cutoff

        self.set_params(lag=lag, dim=dim, var_cutoff=var_cutoff, kinetic_map=kinetic_map, commute_map=commute_map,
                        epsilon=epsilon, reversible=reversible, stride=stride, skip=skip, weights=weights, ncov_max=ncov_max)

    @property
    def model(self):
        if not hasattr(self, '_model') or self._model is None:
            self._model = TICAModelBase()
        return self._model

    def describe(self):
        try:
            dim = self.dimension()
        except RuntimeError:
            dim = self.dim
        return "[TICA, lag = %i; max. output dim. = %i]" % (self._lag, dim)

    def estimate(self, X, **kwargs):
        r"""
        Chunk-based parameterization of TICA. Iterates over all data and estimates
        the mean, covariance and time lagged covariance. Finally, the
        generalized eigenvalue problem is solved to determine
        the independent components.
        """
        return super(TICA, self).estimate(X, **kwargs)

    def partial_fit(self, X):
        """ incrementally update the covariances and mean.

        Parameters
        ----------
        X: array, list of arrays, PyEMMA reader
            input data.

        Notes
        -----
        The projection matrix is first being calculated upon its first access.
        """
        from pyemma.coordinates import source
        iterable = source(X, chunksize=self.chunksize)

        indim = iterable.dimension()
        if not self.dim <= indim:
            raise RuntimeError("requested more output dimensions (%i) than dimension"
                               " of input data (%i)" % (self.dim, indim))
        if self._covar is None:
            self._covar = LaggedCovariance(c00=True, c0t=True, ctt=False, remove_data_mean=True, reversible=self.reversible,
                                           lag=self.lag, bessel=False, stride=self.stride, skip=self.skip,
                                           weights=self.weights, ncov_max=self.ncov_max)
        self._covar.partial_fit(iterable)
        self.model.update_model_params(mean=self._covar.mean,  # TODO: inefficient, fixme
                                        cov=self._covar.C00_,
                                        cov_tau=self._covar.C0t_)

        self._estimated = False

        return self

    def _estimate(self, iterable, **kw):
        covar = LaggedCovariance(c00=True, c0t=True, ctt=False, remove_data_mean=True, reversible=self.reversible,
                                 lag=self.lag, bessel=False, stride=self.stride, skip=self.skip,
                                 weights=self.weights, ncov_max=self.ncov_max)
        indim = iterable.dimension()

        if not self.dim <= indim:
            raise RuntimeError("requested more output dimensions (%i) than dimension"
                               " of input data (%i)" % (self.dim, indim))

        if self._logger_is_active(self._loglevel_DEBUG):
            self.logger.debug("Running TICA with tau=%i; Estimating two covariance matrices"
                              " with dimension (%i, %i)", self._lag, indim, indim)
        covar.estimate(iterable, chunksize=self.chunksize, **kw)
        self.model.update_model_params(mean=covar.mean,
                                        cov=covar.C00_,
                                        cov_tau=covar.C0t_)
        self._diagonalize()

        return self.model

    def _diagonalize(self):
        # diagonalize with low rank approximation
        self.logger.debug("diagonalize Cov and Cov_tau.")
        try:
            eigenvalues, eigenvectors = eig_corr(self.cov, self.cov_tau, self.epsilon, sign_maxelement=True)
        except ZeroRankError:
            raise ZeroRankError('All input features are constant in all time steps. No dimension would be left after dimension reduction.')
        if self.kinetic_map and self.commute_map:
            raise ValueError('Trying to use both kinetic_map and commute_map. Use either or.')
        if self.kinetic_map:  # scale by eigenvalues
            eigenvectors *= eigenvalues[None, :]
        if self.commute_map:  # scale by (regularized) timescales
            timescales = 1-self.lag / np.log(np.abs(eigenvalues))
            # dampen timescales smaller than the lag time, as in section 2.5 of ref. [5]
            regularized_timescales = 0.5 * timescales * np.maximum(np.tanh(np.pi * ((timescales - self.lag) / self.lag) + 1), 0)

            eigenvectors *= np.sqrt(regularized_timescales / 2)
        self.logger.debug("finished diagonalisation.")

        # compute cumulative variance
        cumvar = np.cumsum(np.abs(eigenvalues) ** 2)
        cumvar /= cumvar[-1]

        self.model.update_model_params(cumvar=cumvar,
                                       eigenvalues=eigenvalues,
                                       eigenvectors=eigenvectors)

        self._estimated = True
